import re
import os
import stat
import argparse
import subprocess
import multiprocessing
import shutil
import textwrap
import pprint
import time
import gc
import itertools

import pysam
import numpy as np
import pyranges as pr
import pandas as pd

import handygenome.logutils as logutils
import handygenome.tools as tools
import handygenome.refgenome as refgenome
import handygenome.workflow as workflow
import handygenome.workflow.toolsetup as toolsetup
import handygenome.annotation.readstats as libreadstats
import handygenome.variant.varianthandler as varianthandler
import handygenome.variant.variantplus as variantplus
import handygenome.variant.filter as libfilter
import handygenome.variant.infoformat as infoformat
import handygenome.vcfeditor.indexing as indexing
import handygenome.vcfeditor.split as libsplit
import handygenome.vcfeditor.concat as libconcat
import handygenome.variant.vcfspec as libvcfspec
import handygenome.variant.ponbams as libponbams
import handygenome.bameditor as bameditor
import handygenome.ucscdata as ucscdata
import handygenome.blacklist as blacklist
import handygenome.vcfeditor.misc as vcfmisc


LOGGER_NAME = __name__.split('.')[-1]


def unit_job(
    split_outfile_path,
    fetchregion,
    infile_copy_path, 
    bam_path_list, 
    id_list, 
    pon_bam_path_list, 
    pon_id_list, 
    refver, 
    no_matesearch, 
    countonly, 
    memuse_limit_gb,
    depth_limits,
    mq_limits,
    include_blacklist,
    monitor_interval=10,
):
    # setup manager
    manager = multiprocessing.Manager()
    read_vrspecs = manager.list()
    shareddict = manager.dict()
    shareddict['finished'] = False
    shareddict['outfile_path'] = split_outfile_path
    shareddict['parent_memuse_gb'] = tools.get_rss(mode='total', unit='g')
    logutils.print_timestamp(f"Memory usage: {shareddict['parent_memuse_gb']} GB")

    # setup paramaters
    blacklist_gr = toolsetup.get_blacklist_gr(refver)

    # make subprocess arguments
    subproc_kwargs = {
        'shareddict': shareddict,
        'read_vrspecs': read_vrspecs,

        'infile_copy_path': infile_copy_path, 
        'fetchregion': fetchregion,
        'bam_path_list': bam_path_list, 
        'id_list': id_list, 
        'pon_bam_path_list': pon_bam_path_list, 
        'pon_id_list': pon_id_list, 
        'refver': refver, 
        'no_matesearch': no_matesearch, 
        'countonly': countonly, 
        'memuse_limit_gb': memuse_limit_gb,
        'depth_limits': depth_limits,
        'mq_limits': mq_limits,
        'include_blacklist': include_blacklist,
        'blacklist_gr': blacklist_gr,
    }

    # main
    while True:
        p = multiprocessing.Process(target=unit_job_core, kwargs=subproc_kwargs)
        p.start()
        while True:
            time.sleep(monitor_interval)
            shareddict['parent_memuse_gb'] = tools.get_rss(mode='total', unit='g')
            logutils.print_timestamp(f"Memory usage: {shareddict['parent_memuse_gb']} GB")
            if not p.is_alive():
                break

        if p.exitcode != 0:
            raise Exception(f'The subprocess finished with nonzero exit code.')
        p.close()

        if shareddict['finished']:
            break
        else:
            logutils.print_timestamp(
                f"Beginning next cycle with new output file: {shareddict['outfile_path']}"
            )
            continue


def unit_job_core(
    shareddict,
    read_vrspecs,
    infile_copy_path,
    fetchregion,
    bam_path_list,
    id_list,
    pon_bam_path_list,
    pon_id_list,
    refver,
    no_matesearch,
    countonly,
    memuse_limit_gb,
    depth_limits,
    mq_limits,
    include_blacklist,
    blacklist_gr,
):
    # basic setup
    bam_dict = {
        sampleid: pysam.AlignmentFile(bam_path)
        for sampleid, bam_path in zip(id_list, bam_path_list)
    }
    pon_bam_dict = {
        sampleid: pysam.AlignmentFile(bam_path)
        for sampleid, bam_path in zip(pon_id_list, pon_bam_path_list)
    }
    fasta = refgenome.get_default_fasta(refver)
    chromdict = refgenome.get_default_chromdict(refver)

    # edit vcf header
    in_vcf = pysam.VariantFile(infile_copy_path, 'r')
    new_header = in_vcf.header.copy()
    added_new_samples = update_header(new_header, id_list, pon_id_list)
    out_vcf = pysam.VariantFile(shareddict['outfile_path'], 'wz', header=new_header)

    # loop over variant records
    vr_iterator = itertools.chain.from_iterable(
        vcfmisc.get_vr_fetcher(
            in_vcf, refver, fetchargs[0], fetchargs[1], fetchargs[2],
            respect_refregion=False,
        )
        for fetchargs in fetchregion
    )
    shareddict['finished'] = True
    for vr in vr_iterator:
        vrspec = '\t'.join([vr.contig, str(vr.pos), vr.ref, ','.join(vr.alts)])
        if vrspec in read_vrspecs:
            continue

        read_vrspecs.append(vrspec)
        logutils.print_timestamp(f'Processing {vrspec}')  # for logging

        if added_new_samples:
            new_vr = varianthandler.reheader(vr, new_header)
        else:
            new_vr = vr

        update_new_vr(
            new_vr, fasta, chromdict, bam_dict, pon_bam_dict, no_matesearch, countonly,
            depth_limits, mq_limits, blacklist_gr,
        )
        out_vcf.write(new_vr)

        if shareddict['parent_memuse_gb'] > memuse_limit_gb:
            next_outfile_path = toolsetup.make_next_split_vcf_path(shareddict['outfile_path'])
            shareddict['outfile_path'] = next_outfile_path
            shareddict['finished'] = False
            break

    out_vcf.close()
    in_vcf.close()


##################################


def unit_job_old(
    split_infile_path, 
    split_outfiles_dir, 
    bam_path_list, 
    id_list, 
    pon_bam_path_list, 
    pon_id_list, 
    refver, 
    no_matesearch, 
    countonly, 
    memuse_limit_gb,
    depth_limits,
    mq_limits,
    include_blacklist,
    monitor_interval=10,
):
    # setup manager
    manager = multiprocessing.Manager()
    shareddict = manager.dict()
    shareddict['parent_memuse_gb'] = tools.get_rss(mode='total', unit='g')
    logutils.print_timestamp(f"Memory usage: {shareddict['parent_memuse_gb']} GB")

    # setup paramaters
#    bam_dict = {
#        sampleid: pysam.AlignmentFile(bam_path)
#        for sampleid, bam_path in zip(id_list, bam_path_list)
#    }
#    pon_bam_dict = {
#        sampleid: pysam.AlignmentFile(bam_path)
#        for sampleid, bam_path in zip(pon_id_list, pon_bam_path_list)
#    }
#    fasta = refgenome.get_default_fasta(refver)
#    chromdict = refgenome.get_default_chromdict(refver)

    cytoband_gr = ucscdata.get_cytoband_gr(refver=refver, rename_hg19=True)
    centromere_gr = cytoband_gr[cytoband_gr.Stain == 'acen']
    blacklist_gr = pr.concat([centromere_gr, blacklist.CURATED_BLACKLIST_GRS[refver]])
    blacklist_gr = blacklist_gr.sort().merge()

    # make subprocess arguments
    subproc_kwargs = {
        'split_infile_path': split_infile_path, 
        'split_outfiles_dir': split_outfiles_dir, 
        'bam_path_list': bam_path_list, 
        'id_list': id_list, 
        'pon_bam_path_list': pon_bam_path_list, 
        'pon_id_list': pon_id_list, 
        'refver': refver, 
        'no_matesearch': no_matesearch, 
        'countonly': countonly, 
        'shareddict': shareddict,
        'memuse_limit_gb': memuse_limit_gb,
        'depth_limits': depth_limits,
        'mq_limits': mq_limits,
        'include_blacklist': include_blacklist,
        'blacklist_gr': blacklist_gr,
    }

    # main
    while True:
        p = multiprocessing.Process(target=unit_job_core, kwargs=subproc_kwargs)
        shareddict['next_infile_path'] = None
        p.start()
        while True:
            time.sleep(monitor_interval)
            shareddict['parent_memuse_gb'] = tools.get_rss(mode='total', unit='g')
            logutils.print_timestamp(f"Memory usage: {shareddict['parent_memuse_gb']} GB")
            if not p.is_alive():
                break

        if p.exitcode != 0:
            raise Exception(f'The subprocess finished with nonzero exit code.')
        p.close()

        if shareddict['next_infile_path'] is None:
            break
        else:
            logutils.print_timestamp(f"Next infile created: {shareddict['next_infile_path']}")
            subproc_kwargs['split_infile_path'] = shareddict['next_infile_path']
            continue


def unit_job_core_old(
    split_infile_path, 
    split_outfiles_dir, 
    bam_path_list, 
    id_list, 
    pon_bam_path_list, 
    pon_id_list, 
    refver, 
    no_matesearch, 
    countonly, 
    shareddict,
    memuse_limit_gb,
    depth_limits,
    mq_limits,
    include_blacklist,
    blacklist_gr,
):
    # basic setup
    bam_dict = {
        sampleid: pysam.AlignmentFile(bam_path)
        for sampleid, bam_path in zip(id_list, bam_path_list)
    }
    pon_bam_dict = {
        sampleid: pysam.AlignmentFile(bam_path)
        for sampleid, bam_path in zip(pon_id_list, pon_bam_path_list)
    }
    fasta = refgenome.get_default_fasta(refver)
    chromdict = refgenome.get_default_chromdict(refver)

    split_outfile_path = os.path.join(
        split_outfiles_dir, 
        os.path.basename(split_infile_path),
    )

    # edit vcf header
    in_vcf = pysam.VariantFile(split_infile_path, 'r')
    new_header = in_vcf.header.copy()
    added_new_samples = update_header(new_header, id_list, pon_id_list)
    out_vcf = pysam.VariantFile(split_outfile_path, 'wz', header=new_header)

    # loop over variant records
    vr_iterator = in_vcf.fetch()
    for vr in vr_iterator:
        vrspec = '\t'.join([vr.contig, str(vr.pos), vr.ref, ','.join(vr.alts)])
        logutils.print_timestamp(f'Processing {vrspec}')  # for logging

        if added_new_samples:
            new_vr = varianthandler.reheader(vr, new_header)
        else:
            new_vr = vr

        update_new_vr(
            new_vr, fasta, chromdict, bam_dict, pon_bam_dict, no_matesearch, countonly,
            depth_limits, mq_limits, blacklist_gr,
        )
        out_vcf.write(new_vr)

        if shareddict['parent_memuse_gb'] > memuse_limit_gb:
            #next_split_infile_path = re.sub('\.vcf\.gz$', 'a.vcf.gz', split_infile_path)
            next_split_infile_path = toolsetup.make_next_infile(split_infile_path)
            next_out_vcf = pysam.VariantFile(next_split_infile_path, 'wz', header=in_vcf.header.copy())
            shareddict['next_infile_path'] = next_split_infile_path
            for vr in vr_iterator:
                next_out_vcf.write(vr)
            next_out_vcf.close()
            break

    out_vcf.close()
    in_vcf.close()


def update_header(vcfheader, id_list, pon_id_list):
    libreadstats.ReadStatsSampledict.add_meta(vcfheader)
    added_new_samples = False
    for sampleid in itertools.chain(id_list, pon_id_list):
        if sampleid not in vcfheader.samples:
            added_new_samples = True
            vcfheader.samples.add(sampleid)

    return added_new_samples


def update_new_vr(
    new_vr, fasta, chromdict, bam_dict, pon_bam_dict, no_matesearch, countonly,
    depth_limits, mq_limits, blacklist_gr,
):
    vcfspec = libvcfspec.Vcfspec.from_vr(new_vr)
    if vcfspec.to_gr().intersect(blacklist_gr).empty:
        init_invalid = False
    else:
        init_invalid = True

    readstats_dict = libreadstats.ReadStatsSampledict.from_bam_dict(
        bam_dict, vcfspec, fasta, chromdict,
        rpplist_kwargs={'no_matesearch': no_matesearch},
        countonly=countonly,
        depth_limits=depth_limits, 
        mq_limits=mq_limits,
        init_invalid=init_invalid,
    )
    readstats_dict.update_bams(
        bam_dict=pon_bam_dict,
        rpplist_kwargs={'no_matesearch': no_matesearch},
        countonly=True,
        depth_limits=depth_limits, 
        mq_limits=mq_limits,
        init_invalid=init_invalid,
    )
    readstats_dict.write(new_vr)

    del readstats_dict
    gc.collect()

    # get irrelevant sample ids: these need to be set '.' afterward
    #FORMAT_key = libreadstats.ReadStatsSampledict.annotkey
#    irrelevant_samples = set(
#        x for x in new_vr.samples.keys()
#        if infoformat.check_NA_format(new_vr, x, FORMAT_key)
#    )
#    irrelevant_samples.difference_update(id_list)


    #for sampleid in irrelevant_samples:
    #    infoformat.set_NA_format(new_vr, sampleid, FORMAT_key)


#########################################################


def argument_parser(cmdargs):
    def sanity_check(args):
        if all(
            [
                (args.bamlist is None),
                (args.bamlist_file_path is None),
                (args.pon_cohorts is None),
            ]
        ):
            raise Exception(f'You must use at least one of "--bamlist", "--bamlist-file", or "--pon-cohorts".')

        if (args.bamlist is None) != (args.idlist is None):
            raise Exception(f'--bamlist and --idlist options must be used together.')

        # --bamlist and --idlist length comparison
        if (args.bamlist is not None) and (args.idlist is not None):
            if len(args.bamlist) != len(args.idlist):
                raise Exception(f'The number of --bamlist and --idlist arguments must be the same.')

    def bam_path_processing(args):
        # all information specified by --bamlist/--idlist, --bamlist-file, --pon-cohorts are added
        # handling --bamlist-file
        if args.bamlist_file_path is not None:
            with open(args.bamlist_file_path, 'r') as infile:
                for line in infile:
                    linesp = line.replace('\n', '').split('\t')
                    if len(linesp) != 2:
                        raise Exception(f'The number of tab-seperated fields must be 2 in the bamlist file.')

                    sampleid = linesp[0]
                    bam_path = linesp[1]
                    if not bam_path.endswith('.bam'):
                        raise Exception(
                            f'Bam file path does not end with ".bam" in the '
                            f'bamlist file. Please check if column 1 is '
                            f'sample ID and column 2 is bam file path.'
                        )

                    args.idlist.append(sampleid)
                    args.bamlist.append(bam_path)
        # handling --pon-cohorts
        args.pon_idlist = list()
        args.pon_bamlist = list()
        if args.pon_cohorts is not None:
            for cohort in args.pon_cohorts:
                for sampleid, bam_path in libponbams.PON_BAM_PATHS[args.refver][cohort].items():
                    args.pon_idlist.append(sampleid)
                    args.pon_bamlist.append(bam_path)

    parser_dict = workflow.init_parser(
        description=(
            f'Calculates allele-supporting read count information for each '
            f'input bam, then writes to FORMAT. To specify input bam '
            f'files and sample IDs, you can use any combination of:\n'
            f'1) --bamlist and --idlist ; '
            f'2) --bamlist-file ; '
            f'3) --pon-cohorts\n'
            f'All these bam files will be used for annotation.\n'
            f'Bam files specified with "--pon-cohorts" will be used with '
            f'"countonly" mode, regardless of the usage of "--countonly" argument.'
        )
    )

    # required
    workflow.add_infile_arg(parser_dict['required'], required=True)
    workflow.add_outfile_arg(parser_dict['required'], required=True, must_not_exist='ask')
    workflow.add_refver_arg(parser_dict['required'], required=True)

    # optional
    parser_dict['optional'].add_argument(
        '--bamlist', 
        dest='bamlist', 
        required=False,
        nargs='+', 
        type=workflow.arghandler_infile, 
        default=list(),
        #metavar='<Input bam file path>',
        help=f'One or more input bam file paths separated by whitespaces.',
    )
    parser_dict['optional'].add_argument(
        '--idlist', 
        dest='idlist', 
        required=False,
        nargs='+', 
        default=list(),
        #metavar='<Sample ID>',
        help=(
            f'Sample IDs of the input bam files, in the same order, '
            f'separated by whitespaces.'
        ),
    )
    parser_dict['optional'].add_argument(
        '--bamlist-file', dest='bamlist_file_path', required=False,
        type=workflow.arghandler_infile,
        #metavar='<bam list file path>',
        help=(
            f'A 2-column tab-separated file which contains sample IDs on '
            f'the first column and bam file paths on the second column.'
        ),
    )

    allowed_pon_cohorts = dict()
    for refver, subdict in libponbams.PON_BAM_PATHS_WITHOUT_NAMES.items():
        allowed_pon_cohorts[refver] = list(subdict.keys())
    parser_dict['optional'].add_argument(
        '--pon-cohorts', dest='pon_cohorts', required=False,
        nargs='+',
        #metavar='<PON cohort name>',
        help=(
            f'One or more known Panel Of Normal cohort names. Allowed values:\n'
            f'{textwrap.indent(pprint.pformat(allowed_pon_cohorts), " " * 4)}'
        ),
    )

    parser_dict['optional'].add_argument(
        '--memlimit', dest='memuse_limit_gb', required=False,
        default=4.0,
        type=float,
        #metavar='<memory limit per job>',
        help=f'Maximum memory limit per job in gigabytes.',
    )

    parser_dict['optional'].add_argument(
        '--depth-limits', dest='depth_limits', required=False,
        default=None,
        #type=float,
        #metavar='<depth upper limit>',
        nargs='+',
        help=(
            f'Limits (both inclusive) of valid depth range. '
            f'If depth of certain site is out of this range, ReadStats '
            f'information will be all set as NaN. -1 given as the second value '
            f'is interpreted as positive infinity. '
            f'Default: lower is 0, upper is '
            f'(--depthlimit-cov-ratio value) * (average depth of the bam file)'
        ),
    )
    parser_dict['optional'].add_argument(
        '--depthlimit-cov-ratio', dest='depthlimit_cov_ratio', required=False,
        default=4,
        type=float,
        #metavar='<bam coverage multiplication factor>',
        help=(
            f'Upper depth limit is set as (this value) * (average bam depth).'
            f'If "--depth-limits" option is used, this option is ignored.'
        ),
    )
    parser_dict['optional'].add_argument(
        '--target-bed', dest='target_bed_path', required=False,
        default=None,
        #metavar='<target region bed file>',
        help=(
            f'If this option is used, bam file average depth calculation will '
            f'be based on the region length derived from this file. '
            f'If "--depth-limits" option is used, this option is ignored.'
        ),
    )
    parser_dict['optional'].add_argument(
        '--mq-limits', dest='mq_limits', required=False,
        default=[
            ','.join(str(x) for x in libreadstats.DEFAULT_MQ_LIMITS)
        ],
        #type=float,
        #metavar='<MQ limits>',
        nargs='+',
        help=(
            f'Limits (both inclusive) of valid mapping quality range. '
            f'If mapping quality of certain site is out of this range, ReadStats '
            f'information will be all set as NaN. -1 given as the second value '
            f'is interpreted as positive infinity.'
        ),
    )

    workflow.add_outfmt_arg(parser_dict['optional'], required=False)
    workflow.add_logging_args(parser_dict)
    workflow.add_scheduler_args(parser_dict, default_parallel=1, default_sched='slurm')

    # flag
    workflow.add_rmtmp_arg(parser_dict)
    workflow.add_index_arg(parser_dict)
    parser_dict['flag'].add_argument(
        '--do-matesearch', dest='do_matesearch', action='store_true',
        help=(f'If set, "no_matesearch" argument is set as False when '
              f'creating ReadPlusPairList objects. It can be much slower, '
              f'but mate reads far away from the variant site can be captured.')
    )
    parser_dict['flag'].add_argument(
        '--countonly', dest='countonly', action='store_true',
        help=f'If set, only read count informations is calculated. Suitable for PON data.'
    )
    parser_dict['flag'].add_argument(
        '--include-blacklist', dest='include_blacklist', action='store_true',
        help=(
            f'By default, variants within blacklist region '
            f'(centromere + curated blacklist regions) '
            f'will be annotated as NaN. '
            f'If this option is set, blacklist region variants will be fully annotated.'
        ),
    )

    # main
    args = parser_dict['main'].parse_args(cmdargs)
    sanity_check(args)
    bam_path_processing(args)

    return args


def depth_mq_limit_processing(args):
    def rawarg_handler(rawarg, nsample):
        if len(rawarg) not in (1, nsample):
            raise Exception(f'Invalid "--depth-limits" or "--mq-limits" arguments')

        if len(rawarg) == 1:
            rawarg = rawarg * nsample
            
        result = list()
        for x in rawarg:
            x = x.split(',')
            minval = float(x[0])
            maxval = float(x[1])
            result.append([minval, maxval])

        return result

    # calculate region length
    if args.depth_limits is None:
        if args.target_bed_path is None:
            sample_bam_path = next(itertools.chain(args.bamlist, args.pon_bamlist))
            with pysam.AlignmentFile(sample_bam_path) as in_bam:
                region_length = sum(in_bam.lengths)
        else:
            region_gr = pr.read_bed(args.target_bed_path)
            region_length = region_gr.merge().length

    # modify raw argument
    sampleid_list = list(itertools.chain(args.idlist, args.pon_idlist))
    bampath_list = list(itertools.chain(args.bamlist, args.pon_bamlist))
    nsample = len(sampleid_list)

    mq_limits = dict(zip(sampleid_list, rawarg_handler(args.mq_limits, nsample)))

    if args.depth_limits is None:
        depth_limits = dict()
        for sampleid, bam_path in zip(sampleid_list, bampath_list):
            with pysam.AlignmentFile(bam_path) as in_bam:
                upper = (
                    bameditor.get_average_depth(in_bam, aligned_region_length=region_length) 
                    * args.depthlimit_cov_ratio
                )
            depth_limits[sampleid] = [0, upper]
    else:
        depth_limits = dict(zip(sampleid_list, rawarg_handler(args.depth_limits, nsample)))

#    mq_limits = dict()
#    depth_limits = dict()
#    for sampleid, bam_path in zip(
#        itertools.chain(args.idlist, args.pon_idlist),
#        itertools.chain(args.bamlist, args.pon_bamlist),
#    ):
#        mq_limits[sampleid] = list(args.mq_limits)
#        if args.depth_limits is None:
#            with pysam.AlignmentFile(bam_path) as in_bam:
#                upper = (
#                    bameditor.get_average_depth(in_bam, aligned_region_length=region_length) 
#                    * args.depthlimit_cov_ratio
#                )
#                depth_limits[sampleid] = [0, upper]
#        else:
#            depth_limits[sampleid] = list(args.depth_limits)

    return depth_limits, mq_limits


def write_jobscripts(
    tmpdir_paths, 
    fetchregion_list,
    parallel, 
    infile_copy_path,
    bam_path_list, 
    id_list, 
    pon_bam_path_list, 
    pon_id_list, 
    refver, 
    no_matesearch,
    countonly,
    memuse_limit_gb,
    depth_limits,
    mq_limits,
    include_blacklist,
    jobname_prefix=__name__.split('.')[-1], 
    nproc=1,
):
    jobscript_path_list = list()
    log_path_list = list()
    slurm_log_pf_list = list()
    split_outfile_path_list = list()
    for zidx in tools.zrange(parallel):
        jobscript_path = os.path.join(tmpdir_paths['scripts'], f'{zidx}.sbatch')
        jobscript_path_list.append(jobscript_path)

        bname = os.path.basename(jobscript_path)
        log_path_list.append(
            os.path.join(tmpdir_paths['logs'], bname + '.log')
        )
        slurm_log_pf_list.append(
            os.path.join(tmpdir_paths['slurm_logs'], bname)
        )

        split_outfile_path = os.path.join(tmpdir_paths['split_outfiles'], f'{zidx}.vcf.gz')
        split_outfile_path_list.append(split_outfile_path)

    kwargs_single = {
        'infile_copy_path': infile_copy_path,
        'bam_path_list': bam_path_list,
        'id_list': id_list,
        'pon_bam_path_list': pon_bam_path_list,
        'pon_id_list': pon_id_list,
        'refver': refver,
        'no_matesearch': no_matesearch,
        'countonly': countonly,
        'memuse_limit_gb': memuse_limit_gb,
        'depth_limits': depth_limits,
        'mq_limits': mq_limits,
        'include_blacklist': include_blacklist,
    }
    kwargs_multi = {
        'split_outfile_path': split_outfile_path_list,
        'fetchregion': fetchregion_list,
    }

    toolsetup.write_jobscripts(
        script_path_list=jobscript_path_list,
        log_path_list=log_path_list,
        slurm_log_pf_list=slurm_log_pf_list,
        module_name=__name__,
        unit_job_func_name='unit_job',
        kwargs_single=kwargs_single,
        kwargs_multi=kwargs_multi,
        jobname_prefix=jobname_prefix,
        nproc=nproc,
    )

    return jobscript_path_list


#def write_jobscripts_old(
#    tmpdir_paths, 
#    split_infile_path_list, 
#    bam_path_list, 
#    id_list, 
#    pon_bam_path_list, 
#    pon_id_list, 
#    refver, 
#    no_matesearch,
#    countonly,
#    memuse_limit_gb,
#    depth_limits,
#    mq_limits,
#    include_blacklist,
#    jobname_prefix=__name__.split('.')[-1], 
#    nproc=1,
#):
#    jobscript_path_list = list()
#    log_path_list = list()
#    for zidx in common.zrange(len(split_infile_path_list)):
#        jobscript_path = os.path.join(tmpdir_paths['scripts'], f'{zidx}.sbatch')
#        jobscript_path_list.append(jobscript_path)
#        log_path = os.path.join(tmpdir_paths['logs'], os.path.basename(jobscript_path) + '.log')
#        log_path_list.append(log_path)
#
#    kwargs_single = {
#        'split_outfiles_dir': tmpdir_paths['split_outfiles'],
#        'bam_path_list': bam_path_list,
#        'id_list': id_list,
#        'pon_bam_path_list': pon_bam_path_list,
#        'pon_id_list': pon_id_list,
#        'refver': refver,
#        'no_matesearch': no_matesearch,
#        'countonly': countonly,
#        'memuse_limit_gb': memuse_limit_gb,
#        'depth_limits': depth_limits,
#        'mq_limits': mq_limits,
#        'include_blacklist': include_blacklist,
#    }
#    kwargs_multi = {
#        'split_infile_path': split_infile_path_list,
#    }
#
#    toolsetup.write_jobscripts(
#        script_path_list=jobscript_path_list,
#        log_path_list=log_path_list,
#        module_name=__name__,
#        unit_job_func_name='unit_job',
#        kwargs_single=kwargs_single,
#        kwargs_multi=kwargs_multi,
#        jobname_prefix=jobname_prefix,
#        nproc=nproc,
#    )
#
#    return jobscript_path_list


def main(cmdargs):
    args = argument_parser(cmdargs)

    # make tmpdir tree
    tmpdir_paths = workflow.get_tmpdir_paths(
        ['scripts', 'logs', 'slurm_logs', 'split_outfiles'],
        prefix = (
            os.path.basename(args.infile_path)
            + '_'
            + __name__.split('.')[-1]
            + '_'
        ),
        where=os.path.dirname(args.infile_path),
    )

    # setup logger
    logger = toolsetup.setup_logger(
        args=args,
        tmpdir_root=tmpdir_paths['root'],
        with_genlog=True,
    )
    logger.info('Beginning')

    # postprocess depth and mq limits
    logger.info(f'Processing depth and MQ limits (may take a while calculating average depths of bam files)')
    depth_limits, mq_limits = depth_mq_limit_processing(args)
    logger.info(f'Depth limits: {depth_limits}')
    logger.info(f'MQ limits: {mq_limits}')

    # make infile copy
    infile_copy_path, infile_copy_index_path = toolsetup.make_infile_copy(
        args.infile_path, tmpdir_paths['root'], logger,
    )

    # make split fetchregions
    logger.info(f'Getting split fetch regions of input VCF')
    fetchregion_list = vcfmisc.get_vcf_fetchregions(
        infile_copy_path, args.parallel, args.refver, verbose=False,
    )
    #logger.info(f'Split fetch regions: {fetchregion_list}')

    # make job scripts and run
    logger.info(f'Creating split job scripts')
    jobscript_path_list = write_jobscripts(
        tmpdir_paths, 
        fetchregion_list,
        args.parallel,
        infile_copy_path,
        args.bamlist, 
        args.idlist, 
        args.pon_bamlist, 
        args.pon_idlist, 
        args.refver, 
        (not args.do_matesearch),
        args.countonly,
        args.memuse_limit_gb,
        depth_limits,
        mq_limits,
        args.include_blacklist,
    )
    logger.info('Running annotation jobs for each split file')
    workflow.run_jobs(
        jobscript_path_list, 
        sched=args.sched, 
        intv_check=args.intv_check, 
        intv_submit=args.intv_submit, 
        max_submit=args.max_submit, 
        logger=logger, 
        log_dir=tmpdir_paths['logs'],
        job_status_logpath=os.path.join(tmpdir_paths['root'], 'job_status_log.gz'),
        raise_on_failure=True,
    )

    # concatenates split files
    logger.info('Merging split files')
    outfile_path_list = tools.listdir(tmpdir_paths['split_outfiles'])
    libconcat.main(
        infile_path_list=outfile_path_list,
        outfile_path=args.outfile_path, 
        mode_pysam=args.mode_pysam,
        outfile_must_not_exist='no',
    )

    # rmtmp, index
    if not args.donot_rm_tmp:
        shutil.rmtree(tmpdir_paths['root'])
    if not args.donot_index:
        indexing.index_vcf(args.outfile_path)

    logger.info('All successfully finished')


#def main_old(cmdargs):
#    args = argument_parser(cmdargs)
#
#    # postprocess depth and mq limits
#    logutils.print_timestamp(f'Processing depth and MQ limits (may take a while calculating average depths of bam files)')
#    depth_limits, mq_limits = depth_mq_limit_processing(args)
#    logutils.print_timestamp(f'Depth limits: {depth_limits}')
#    logutils.print_timestamp(f'MQ limits: {mq_limits}')
#
#    # make tmpdir tree
#    tmpdir_paths = workflow.get_tmpdir_paths(
#        ['scripts', 'logs', 'split_infiles', 'split_outfiles'],
#        prefix = (
#            os.path.basename(args.infile_path)
#            + '_'
#            + __name__.split('.')[-1]
#            + '_'
#        ),
#        where=os.path.dirname(args.infile_path),
#    )
#
#    # setup logger
#    logger = toolsetup.setup_logger(args=args,
#                                    tmpdir_root=tmpdir_paths['root'],
#                                    with_genlog=True)
#    logger.info('Beginning')
#
#    # split the input file
#    logger.info('Splitting the input file')
#    split_infile_path_list = libsplit.main(vcf_path=args.infile_path, 
#                                        outdir=tmpdir_paths['split_infiles'], 
#                                        n_file=args.parallel, 
#                                        mode_bcftools='z', prefix='', 
#                                        suffix='.vcf.gz')
#
#    # make job scripts and run
#    jobscript_path_list = write_jobscripts(
#        tmpdir_paths, 
#        split_infile_path_list, 
#        args.bamlist, 
#        args.idlist, 
#        args.pon_bamlist, 
#        args.pon_idlist, 
#        args.refver, 
#        (not args.do_matesearch),
#        args.countonly,
#        args.memuse_limit_gb,
#        depth_limits,
#        mq_limits,
#        args.include_blacklist,
#    )
#    logger.info('Running annotation jobs for each split file')
#    workflow.run_jobs(
#        jobscript_path_list, 
#        sched=args.sched, 
#        intv_check=args.intv_check, 
#        intv_submit=args.intv_submit, 
#        max_submit=args.max_submit, 
#        logger=logger, 
#        log_dir=tmpdir_paths['logs'],
#        job_status_logpath=os.path.join(tmpdir_paths['root'], 'job_status_log.gz'),
#        raise_on_failure=True,
#    )
#
#    # concatenates split files
#    logger.info('Merging split files')
#    outfile_path_list = common.listdir(tmpdir_paths['split_outfiles'])
#    print(outfile_path_list)
#    libconcat.main(
#        infile_path_list=outfile_path_list,
#        outfile_path=args.outfile_path, 
#        mode_pysam=args.mode_pysam,
#        outfile_must_not_exist='no',
#    )
#
#    # rmtmp, index
#    if not args.donot_rm_tmp:
#        shutil.rmtree(tmpdir_paths['root'])
#    if not args.donot_index:
#        indexing.index_vcf(args.outfile_path)
#
#    logger.info('All successfully finished')



"""Multiple Sequence Alignment"""

import os
import tempfile
import subprocess
import gzip
import collections
import re
import itertools

import Bio.Seq
import Bio.SeqIO
import Bio.SeqRecord
import Bio.AlignIO
import Bio.Align
import Bio.Align.AlignInfo
import numpy as np
import pysam

import handygenome.tools as tools
import handygenome.align.alignhandler as alignhandler


GAP_CHAR = '-'
MUSCLE_PATH = '/home/users/pjh/tools/miniconda/210821/miniconda3/envs/genome_v5/bin/muscle'
PAT_ALIGN = re.compile('^([^-]+)((-+)(([^-])(.*))?)?$')


#################################################################################
# custom-made msa (targeted for softclip segments with identical anchor points) #
#################################################################################

def msa_with_target(target, alignments, target_seqrec=None, query_seqrecs=None):
    # arg handling
    assert (target_seqrec is None) == (query_seqrecs is None)
    seqrec_given = (target_seqrec is not None)

    # make target gaps
    target_gaps = summarize_target_gaps(alignments)

    # make msa
    if seqrec_given:
        input_seqrecs = list()
        input_seqrecs.append(
            add_gaps_to_target(target_seqrec, target_gaps)
        )
        for aln, query_rec in zip(alignments, query_seqrecs):
            input_seqrecs.append(
                add_gaps_to_query(aln, target_gaps, query_seqrec=query_rec)
            )
    else:
        padded_seqs = list()
        padded_seqs.append(
            add_gaps_to_target(target, target_gaps)
        )
        for aln in alignments:
            padded_seqs.append(
                add_gaps_to_query(aln, target_gaps)
            )
        input_seqrecs = [Bio.SeqRecord.SeqRecord(x) for x in padded_seqs]

    msa = Bio.Align.MultipleSeqAlignment(input_seqrecs)

    # make consensus
    consensus = str(Bio.Align.AlignInfo.SummaryInfo(msa).dumb_consensus())

    return msa, consensus


def summarize_target_gaps(aln_list):
    """Make a summary of target gaps in each alignment
    Returns: dict 
        keys: pos0 of the base on the right side of the gap. May include a value equal to "target length"
        values: gap length
    """
    target_gaps = dict()
    for aln in aln_list:
        walks = alignhandler.get_walks(aln)
        for walk in walks:
            if walk.check_queryonly():
                gap_right_border = walk.target.stop
                gap_length = len(walk.query)
                target_gaps.setdefault(gap_right_border, 0)
                target_gaps[gap_right_border] = max(target_gaps[gap_right_border], gap_length)

    return target_gaps


def add_gaps_to_target(target, target_gaps, gap_char=GAP_CHAR):
    if isinstance(target, str):
        return add_gaps_to_target_string(target, target_gaps, gap_char=gap_char)
    elif isinstance(target, Bio.SeqRecord.SeqRecord):
        padded_target = add_gaps_to_target_string(str(target.seq), target_gaps, gap_char=gap_char)
        target.seq = Bio.Seq.Seq(padded_target)
        #new_seqrec = Bio.SeqRecord.SeqRecord(
        #    padded_target, 
        #    annotations=target.annotations,
        #)
        return target
    else:
        raise Exception(f'"target" argument must be either str or Bio.SeqRecord.SeqRecord')


def add_gaps_to_target_string(target, target_gaps, gap_char=GAP_CHAR):
    """Makes a modified target sequence with gap added as '-'"""
    target_list = list(target)
    target_end_idx = len(target)

    for gap_end, gap_len in target_gaps.items():
        if gap_end == target_end_idx:
            target_list.append(gap_char * gap_len)
        else:
            target_list[gap_end] = (gap_char * gap_len) + target_list[gap_end]

    return ''.join(target_list)


def add_gaps_to_query(alignment, target_gaps, query_seqrec=None, overlapping_gap_to_right=True, gap_char=GAP_CHAR):
    if query_seqrec is None:
        return add_gaps_to_query_string(alignment, target_gaps, overlapping_gap_to_right=overlapping_gap_to_right, gap_char=gap_char)
    else:
        padded_query = add_gaps_to_query_string(alignment, target_gaps, overlapping_gap_to_right=overlapping_gap_to_right, gap_char=gap_char)
        query_seqrec.seq = Bio.Seq.Seq(padded_query)
        #new_seqrec = Bio.SeqRecord.SeqRecord(padded_query, annotations=query_seqrec.annotations)
        return query_seqrec


def add_gaps_to_query_string(alignment, target_gaps, overlapping_gap_to_right=True, gap_char=GAP_CHAR):
    """Given an alignment between target and query, returns a modified query sequence with gap added as '-'"""
    target_indices = alignment.indices[0]
    target_indices_list = list(target_indices)

    original_gap_info = list()
    for is_gap, subiter in itertools.groupby(
        enumerate(target_indices), key=(lambda x: x[1] == -1)
    ):
        if is_gap:
            subiter = list(subiter)
            original_gapend_alnidx = subiter[-1][0] + 1
            original_gaplen = len(subiter)
            original_gap_info.append((original_gapend_alnidx, original_gaplen))

    raw_query_list = list(alignment[1])
    raw_query_list.append('')
    for (new_gapend_idx, new_gaplen) in target_gaps.items():
        if new_gapend_idx == len(alignment.target):
            new_gapend_alnidx = len(target_indices)
        else:
            new_gapend_alnidx = target_indices_list.index(new_gapend_idx)

        ovlp_original_gaps = [
            (original_gapend_alnidx, original_gaplen) 
            for (original_gapend_alnidx, original_gaplen) in original_gap_info 
            if (new_gapend_alnidx == original_gapend_alnidx)
        ]
        assert len(ovlp_original_gaps) <= 1

        if len(ovlp_original_gaps) == 0:
            query_insertion_idx = new_gapend_alnidx
            added_gaplen = new_gaplen
        else:
            original_gapend_alnidx, original_gaplen = ovlp_original_gaps[0]

            if overlapping_gap_to_right:
                query_insertion_idx = new_gapend_alnidx
            else:
                query_insertion_idx = new_gapend_alnidx - original_gaplen

            added_gaplen = new_gaplen - original_gaplen
            assert added_gaplen >= 0    

        raw_query_list[query_insertion_idx] = (gap_char * added_gaplen) + raw_query_list[query_insertion_idx]

    return ''.join(raw_query_list)

##########################################################




############################################################
############################################################
############################################################
############################################################
############################################################

# old ones #

    
class MultipleSequenceAlignment:
    def __init__(self, seqs, names=None):
        assert len(set(len(x) for x in seqs)) == 1
        
        shape = (len(seqs), len(seqs[0]), 1)
        self.array = np.reshape([list(x) for x in seqs], shape)
        if names is None:
            self.names = None
        else:
            self.names = list(names)

    def __repr__(self):
        sep = ' ' * 4

        lines = list()
        idx_width = len(str(self.array.shape[0] - 1))
        for idx in range(self.array.shape[0]):
            seq = ''.join(self.array[idx].ravel())
            zidx = f'{idx:{idx_width}d}'
            if self.names is None:
                line = sep.join([zidx, seq])
            else:
                name = self.names[idx]
                line = sep.join([zidx, seq, name])
            lines.append(line)

        return '\n'.join(lines)

    def select_col(self, col_idx, rows_incl=None, rows_excl=None):
        if rows_incl is not None:
            row_selector = np.repeat(False, self.array.shape[0])
            row_selector[rows_incl] = True
        elif rows_excl is not None:
            row_selector = np.repeat(True, self.array.shape[0])
            row_selector[rows_excl] = False
        else:
            row_selector = np.repeat(True, self.array.shape[0])

        return self.array[row_selector, col_idx, 0]

    def get_consensus_col(self, col_idx, threshold=0.5, ambiguous='X'):
        col = self.array[:, col_idx, 0]
        col = col[col != '-']
        if col.shape[0] == 0:
            result = ('-')
        else:
            counter = collections.Counter(col)
            max_val = max(counter.values())
            max_bases = [x for x in counter.keys() if counter[x] == max_val]
            if len(max_bases) == 1:
                max_base = max_bases[0]
                total = sum(counter.values())
                prop = counter[max_base] / total
                if prop >= threshold:
                    result = max_base
                else:
                    result = ambiguous
            else:
                result = ambiguous

        return result

    def get_consensus(self, threshold=0.5, ambiguous='X'):
        result = list()
        for col_idx in range(self.array.shape[1]):
            result_col = self.get_consensus_col(col_idx, threshold=threshold, ambiguous=ambiguous)
            result.append(result_col)

        return ''.join(result)

    def get_consensus_biopython(self):
        seqr_list = (Bio.SeqRecord.SeqRecord(''.join(self.array[idx, :, 0]))
                     for idx in range(self.array.shape[0]))
        msa_bio = Bio.Align.MultipleSeqAlignment(seqr_list)
        summ = Bio.Align.AlignInfo.SummaryInfo(msa_bio)

        return str(summ.dumb_consensus())


def modify_msa_gapclose(msa, feasibility_threshold=0.5):
    def check_feasible(msa, row_idx, mat, feasibility_threshold):
        if mat.group(2) is None or mat.group(4) is None:
            return False

        moved_base = mat.group(5)
        col_idx = mat.start(2)
        col_bases = msa.select_col(col_idx, rows_excl=[row_idx])
        base_counter = collections.Counter(col_bases[col_bases != '-'])
        
        numerator = base_counter[moved_base]
        denominator = sum(base_counter.values())
        if denominator == 0:
            return False
        else:
            fraction = numerator / denominator
            return (fraction >= feasibility_threshold)

    def modify_row(msa, row_idx, feasibility_threshold):
        while True:
            mat = PAT_ALIGN.match(''.join(msa.array[row_idx, :, 0]))
            if check_feasible(msa, row_idx, mat, feasibility_threshold):
                new_seq = (mat.group(1) 
                           + mat.group(5) 
                           + mat.group(3) 
                           + mat.group(6))
                msa.array[row_idx, :, 0] = list(new_seq)
                continue
            else:
                break

    # main
    for idx in range(msa.array.shape[0]):
        modify_row(msa, idx, feasibility_threshold)


def fasta_to_msa(fasta_path):
    seqs = list()
    names = list()

    with tools.openfile(fasta_path) as infile:
        for seqrec in Bio.SeqIO.parse(infile, 'fasta'):
            seqs.append(str(seqrec.seq))
            names.append(seqrec.id)

    msa = MultipleSequenceAlignment(seqs=seqs, names=names)

    return msa


def write_fasta(fname, seqlist, seqnamelist=None):
    if fname.endswith('.gz'):
        f = gzip.open(fname, 'wt')
    else:
        f = open(fname, 'wt')

    for seq, seqname in zip(seqlist, seqnamelist):
        f.write('>' + seqname + '\n')
        f.write(seq + '\n')

    f.close()


def run_muscle(infile_path, outfile_path):
    args = [MUSCLE_PATH, '-align', infile_path, '-output', outfile_path]
    p = subprocess.run(args, capture_output=True, text=True, check=True)
    return p


def align_with_muscle(seqlist, seqnamelist=None, biopython=False):
    if seqnamelist is None:
        seqnamelist = [f'seq{x}' for x in range(len(seqlist))]

    with tempfile.TemporaryDirectory(dir=os.getcwd()) as tmpdir:
        infile_path = os.path.join(tmpdir, 'input.fasta')
        outfile_path = os.path.join(tmpdir, 'output.afa')
        write_fasta(infile_path, seqlist, seqnamelist=seqnamelist)
        p = run_muscle(infile_path, outfile_path)

        if biopython:
            msa = Bio.AlignIO.read(outfile_path, 'fasta')
        else:
            msa = fasta_to_msa(outfile_path)

    return msa

                



import colorsys

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt



def lightness_spectrum(color, n, l_start=0.3, l_end=0.7):
    color_hls = colorsys.rgb_to_hls(*mpl.colors.to_rgb(color))
    return [
        colorsys.hls_to_rgb(color_hls[0], x, color_hls[2]) 
        for x in np.linspace(l_start, l_end, n + 2, endpoint=True)[1:-1]
    ]

    



import sys
import logging
import datetime


# timestamp

def get_datestring():
    """Returns a string like '2021-12-06 11:49:55'"""
    return str(datetime.datetime.now()).split('.')[0]


def get_timestamp():
    """Returns a string like 'KST 2021-12-06 11:51:36'"""
    dt = datetime.datetime.now().astimezone()
    return f'{str(dt.tzinfo)} {str(dt).split(".")[0]}'


def print_err(*args, stderr=True, files=None, **kwargs):
    """Args:
        stderr: (Bool) Whether to write to stderr
        files: A list of file paths to which message is written
    """
    if stderr:
        print(*args, file=sys.stderr, flush=True, **kwargs)

    if files is not None:
        for fname in files:
            with open(fname, 'a') as f:
                print(*args, file=f, flush=True, **kwargs)


def print_timestamp(*args, **kwargs):
    print_err(f'[{get_timestamp()}]', *args, **kwargs)


printlog = print_timestamp


# line number logging

def iter_lineno_logging(line_iterator, logger, logging_lineno, msgfunc=None):
    if msgfunc is None:
        def msgfunc(NR):
            return f'Processing {NR:,}th line'

    if logging_lineno is None:
        for line in line_iterator:
            yield line
    else:
        NR = 0
        for line in line_iterator:
            NR += 1
            if NR % logging_lineno == 0:
                logger.info(msgfunc(NR))
            yield line


# making custom logger instance


DEFAULT_DATEFMT = '%Z %Y-%m-%d %H:%M:%S' # KST 2022-03-23 22:12:34
DEFAULT_LOG_FORMATTERS = {
    'without_name': logging.Formatter(
        fmt='[%(asctime)s %(levelname)s] %(message)s', 
        datefmt=DEFAULT_DATEFMT,
    ),
    'with_name': logging.Formatter(
        fmt='[%(asctime)s %(levelname)s] %(name)s: %(message)s', 
        datefmt=DEFAULT_DATEFMT,
    ),
}

def get_logger(
    name=None, 
    formatter=None, 
    level='info', 
    stderr=True, 
    filenames=None, 
    append=False,
):
    if name is None:
        #name = str(uuid.uuid4())
        name = __name__.split('.')[-1]

    if formatter is None:
        formatter = DEFAULT_LOG_FORMATTERS['with_name']

    loglevel = getattr(logging, level.upper())

    logger = logging.getLogger(name)
    logger.setLevel(loglevel)
    logger.propagate = False

    if stderr:
        sh = logging.StreamHandler()
        sh.setLevel(loglevel)
        sh.setFormatter(formatter)
        logger.addHandler(sh)

    if filenames is not None:
        assert isinstance(filenames, (tuple, list))
        for fname in filenames:
            fh = logging.FileHandler(fname, mode=('a' if append else 'w'))
            fh.setLevel(loglevel)
            fh.setFormatter(formatter)
            logger.addHandler(fh)

    return logger


# funclogger

def make_funclogger(level, name):
    formatter = logging.Formatter(
        #fmt=(
        #    '[%(asctime)s.%(msecs)03d %(levelname)s] '
        #    '%(module)s.%(funcName) line %(lineno)d: %(message)s'
        #), 
        fmt=(
            '[%(asctime)s.%(msecs)s %(levelname)s] '
            '%(module)s.%(funcName) line %(lineno)s: %(message)s'
        ), 
        datefmt='%Z %Y-%m-%d %H:%M:%S',
    )
    return get_logger(
        level=level, 
        name=name, 
        formatter=formatter,
        stderr=True, 
        filenames=None, 
        append=False,
    )


FUNCLOGGER_DEBUG = make_funclogger(level='debug', name='FUNCLOGGER_DEBUG')
FUNCLOGGER_INFO = make_funclogger(level='info', name='FUNCLOGGER_INFO')


def get_funclogger(verbose):
    if verbose:
        return FUNCLOGGER_DEBUG
    else:
        return FUNCLOGGER_INFO


def printlog_funclogger(msg):
    FUNCLOGGER_INFO.info(msg)




import os
import argparse
import subprocess
import tempfile

import pysam

import handygenome.refgenome as refgenome
import handygenome.interval as libinterval
import handygenome.workflow as workflow
import handygenome.workflow.toolsetup as toolsetup
import handygenome.gatk as libgatk


PROGNAME = __name__.split(".")[-1]


def unit_job(split_outfile_path, fasta_path, region_path, padded_region_path, mbq):
    # run HaplotypeCaller
    tmpdir = tempfile.TemporaryDirectory(where=os.getcwd())
    split_outfile_path_prefilter = split_outfile_path + ".prefilter"
    libgatk.run_haplotypecaller(
        fasta_path=fasta_path,
        infile_path=infile_path,
        outfile_path=split_outfile_path_prefilter,
        tmpdir=tmpdir,
        incl_bed_path=padded_region_path,
        mbq=mbq,
    )

    # filter
    region_intvlist = libinterval.IntervalList.from_bed(region_path)
    in_vcf = pysam.VariantFile(split_outfile_path_prefilter, "r")
    out_vcf = pysam.VariantFile(split_outfile_path, "wz", header=in_vcf.header.copy())

    for vr in in_vcf.fetch():
        if region_intvlist.includes_vr(vr):
            out_vcf.write(vr)


def argument_parser(cmdargs):
    parser_dict = workflow.init_parser(
        description=(f"Runs gatk HaplotypeCaller with N parallel jobs.")
    )
    # required
    workflow.add_infile_arg(
        parser_dict["required"], required=True, help=f"Input bam file path."
    )
    workflow.add_outfile_arg(
        parser_dict["required"], required=True, must_not_exist="ask"
    )
    workflow.add_fasta_arg(parser_dict["required"], required=True)
    # optional
    workflow.add_outfmt_arg(parser_dict["optional"], required=False)
    workflow.add_incl_region_arg(parser_dict["optional"], required=False)
    workflow.add_excl_region_arg(parser_dict["optional"], required=False)
    workflow.add_logging_args(parser_dict)
    workflow.add_scheduler_args(parser_dict, default_parallel=1, default_sched="slurm")
    # flag
    workflow.add_rmtmp_arg(parser_dict)
    workflow.add_index_arg(parser_dict)
    # main
    args = parser_dict["main"].parse_args(cmdargs)
    return args


def make_tmpdir(infile_path):
    tmpdir_paths = workflow.get_tmpdir_paths(
        ["scripts", "logs", "regions", "split_results"],
        prefix=(os.path.basename(infile_path) + "_" + PROGNAME),
        where=os.path.dirname(infile_path),
    )

    return tmpdir_paths


def main(cmdargs):
    args = argument_parser(cmdargs)

    # make temporary directory tree
    tmpdir_paths = make_tmpdir(args.infile_path)

    # setup logger
    logger = toolsetup.setup_logger(
        args=args, tmpdir_root=tmpdir_paths["root"], with_genlog=True
    )
    logger.info("Beginning")

    # setup other parameters
    chromdict = refgenome.Chromdict.from_fasta_path(args.fasta_path)

    # make split intervals and write bed files
    toolsetup.handle_region_args(
        chromdict, 
        incl_bed_path=args.incl_bed_path,
        excl_bed_path=args.excl_bed_path,
        num_split=args.parallel,
        regionfiles_dir=tmpdir_paths["regions"],
    )




import sys
import collections
import itertools
import functools
import logging
import inspect
import random
import uuid

import pysam
import Bio.Align
import numpy as np
import pandas as pd
import pyranges as pr

import handygenome.refgenome as refgenome
import handygenome.workflow as workflow
import handygenome.variant.vcfspec as libvcfspec
import handygenome.read.pileup as libpileup
import handygenome.align.alignhandler as alignhandler
import handygenome.bameditor as bameditor
import handygenome.read.readhandler as readhandler
import handygenome.read.readplus as readplus
import handygenome.align.realign.base as realign_base


class MultisampleRealignerPileup(realign_base.RealignerPileupBase):
    def __init__(
        self, 
        pileup_dict, 
        refver=None,
        fasta=None, 
        aligner=realign_base.DEFAULT_ALIGNER,
        verbose=False, 
        logger=None,
        **kwargs,
    ):
        """Args:
            pileup_dict: keys - sampleid; values - Pileup object
        """
        # set params
        self.pileup_dict = pileup_dict
        self.refver, self.fasta = realign_base.RealignerPileupBase.refver_fasta_arghandler(refver, fasta)
        self.aligner = aligner
        self.verbose, self.logger = realign_base.RealignerPileupBase.logger_arghandler(verbose, logger)
        self.params = realign_base.parse_rpileup_kwargs(**kwargs)

        # setup data structure
        self.set_df()
        self.set_row_spec_groups()
        self.save_subseq_alignments()
        self.allocate_subseq_alignments()
        self.set_row_spec_groups_bysample()

    @property
    def first_pileup(self):
        return next(iter(self.pileup_dict.values()))

    @property
    def chrom(self):
        return self.first_pileup.chrom

    @property
    def start0(self):
        return self.first_pileup.start0

    @property
    def end0(self):
        return self.first_pileup.end0

    @property
    def range0(self):
        return self.first_pileup.range0

    def set_df(self):
        assert len(set(x.range0 for x in self.pileup_dict.values())) == 1, f'Genomic ranges of pileup objects are different.'

        self.df = pd.concat(
            {key: val.seq_df for key, val in self.pileup_dict.items()},
            names=['SampleID', 'ReadUID'],
        )

    def row_spec_getter(self, key):
        return self.pileup_dict[key[0]].row_specs[key[1]]

    def set_row_spec_groups(self):
        # initialize row_spec_groups
        self.row_spec_groups = dict()
        candidates = dict()
        for superseq_sampleid, sub_pileup in self.pileup_dict.items():
            for superseq_readuid, groupinfo in sub_pileup.row_spec_groups.items():
                superseq_key = (superseq_sampleid, superseq_readuid)
                supserseq_row_spec = sub_pileup.row_specs[superseq_readuid]
                candidates.setdefault(supserseq_row_spec['seq'], list())
                candidates[supserseq_row_spec['seq']].append((superseq_key, supserseq_row_spec))

        for key, val in candidates.items():
            selected_superseq_key = max(
                val, 
                key=(lambda x: (x[1]['span_length'], -x[1]['cliplen'])),
            )[0]
            self.row_spec_groups[selected_superseq_key] = {
                'subseq_rowids': list(),
                'subseq_hits': 0,
            }

        # add subseq entries
        for subseq_sampleid, sub_pileup in self.pileup_dict.items():
            for subseq_readuid, subseq_row_spec in sub_pileup.row_specs.items():
                superseq_candidates = list()
                for superseq_key in self.row_spec_groups.keys():
                    superseq_sampleid, superseq_readuid = superseq_key
                    superseq_row_spec = self.pileup_dict[superseq_sampleid].row_specs[superseq_readuid]
                    if self.row_spec_matcher(subseq_row_spec, superseq_row_spec):
                        superseq_candidates.append(superseq_key)
                        
                if len(superseq_candidates) == 0:
                    raise Exception(f'Subseq row_spec does not match with any of superseq row_specs')
                else:
                    superseq_key = self.superseq_candidates_tiebreak(superseq_candidates, self.row_spec_getter, self.row_spec_groups)
                    groupinfo = self.row_spec_groups[superseq_key]
                    groupinfo['subseq_rowids'].append((subseq_sampleid, subseq_readuid))
                    groupinfo['subseq_hits'] += 1

        # discard 0-hit groups
        no_hits = [
            subseq_key for subseq_key, groupinfo in self.row_spec_groups.items()
            if groupinfo['subseq_hits'] == 0
        ]
        for subseq_key in no_hits:
            del self.row_spec_groups[subseq_key]

        self.decorate_groupinfo(self.row_spec_groups, self.read_getter)

    def read_getter(self, key):
        return self.pileup_dict[key[0]].read_store[key[1]]

    def save_subseq_alignments(self):
        self._subseq_alignments = dict()
        for superseq_key, groupinfo in self.row_spec_groups.items():
            sup_to_ref_aln = self.pileup_dict[superseq_key[0]]._superseq_alignments[superseq_key[1]]
            for subseq_key in groupinfo['subseq_rowids']:
                subseq_row_spec = self.pileup_dict[subseq_key[0]].row_specs[subseq_key[1]]
                try:
                    self._subseq_alignments[subseq_key] = self.align_subseq_to_ref(subseq_row_spec, sup_to_ref_aln)
                except:
                    print(superseq_key)
                    print(subseq_key)
                    raise

    def allocate_subseq_alignments(self):
        subseq_alignments_bysample = {sampleid: dict() for sampleid in self.pileup_dict.keys()}
        for subseq_key, aln in self._subseq_alignments.items():
            subseq_alignments_bysample[subseq_key[0]][subseq_key[1]] = aln

        for sampleid, sub_pileup in self.pileup_dict.items():
            sub_pileup._subseq_alignments = subseq_alignments_bysample[sampleid]

    def set_row_spec_groups_bysample(self):
        self.row_spec_groups_bysample = {sampleid: dict() for sampleid in self.pileup_dict.keys()}
        for sampleid, subgroups in self.row_spec_groups_bysample.items():
            for superseq_key, groupinfo in self.row_spec_groups.items():
                subgroups[superseq_key] = {
                    'subseq_rowids': list(),
                    'subseq_hits': 0,
                }
            
        for superseq_key, groupinfo in self.row_spec_groups.items():
            for subseq_key in groupinfo['subseq_rowids']:
                subgroups = self.row_spec_groups_bysample[subseq_key[0]]
                subgroups[superseq_key]['subseq_rowids'].append(subseq_key)
                subgroups[superseq_key]['subseq_hits'] += 1

        for sampleid, subgroups in self.row_spec_groups_bysample.items():
            self.decorate_groupinfo(subgroups, self.read_getter)

    def iter_contig_vcfspecs(self, subseq_portion_threshold, MQ_threshold, threshold_bysample=True, verbose=False):
        if threshold_bysample:
            for superseq_key, groupinfo in sorted(
                self.row_spec_groups.items(),
                key=(lambda x: x[1]['subseq_hits']),
                reverse=True,
            ):
                if any(
                    (subgroups[superseq_key]['subseq_hits_portion'] >= subseq_portion_threshold) and
                    (subgroups[superseq_key]['mean_MQ'] >= MQ_threshold)
                    for subgroups in self.row_spec_groups_bysample.values()
                ):
                    if verbose:
                        print('vcfspec:', self.pileup_dict[superseq_key[0]].contig_vcfspecs[superseq_key[1]])
                        print('vcfspec components:', self.pileup_dict[superseq_key[0]].contig_vcfspecs[superseq_key[1]].components)
                        for sampleid, subgroups in self.row_spec_groups_bysample.items():
                            print('sample ID:', sampleid, 'portion:', subgroups[superseq_key]['subseq_hits_portion'], 'meanMQ:', subgroups[superseq_key]['mean_MQ'])
                        print()

                    yield superseq_key, self.pileup_dict[superseq_key[0]].contig_vcfspecs[superseq_key[1]]
        else:
            for superseq_key, groupinfo in sorted(
                self.row_spec_groups.items(),
                key=(lambda x: x[1]['subseq_hits']),
                reverse=True,
            ):
                if (
                    (groupinfo['subseq_hits_portion'] >= subseq_portion_threshold) and
                    (groupinfo['mean_MQ'] >= MQ_threshold)
                ):
                    if verbose:
                        print('vcfspec:', self.pileup_dict[superseq_key[0]].contig_vcfspecs[superseq_key[1]])
                        print('vcfspec components:', self.pileup_dict[superseq_key[0]].contig_vcfspecs[superseq_key[1]].components)
                        print('portion:', groupinfo['subseq_hits_portion'], 'meanMQ:', groupinfo['mean_MQ'])
                        print()

                    yield superseq_key, self.pileup_dict[superseq_key[0]].contig_vcfspecs[superseq_key[1]]

    # get_result_vcfspec BEGIN #

    def get_result_vcfspec_recalc_hits(self, vaf_cutoff=0.01, verbose=False):
        candidate_monoalts = list()
        for superseq_key in self.row_spec_groups.keys():
            contig_vcfspec = self.pileup_dict[superseq_key[0]].contig_vcfspecs[superseq_key[1]]
            if contig_vcfspec is not None:
                candidate_monoalts.append(contig_vcfspec.normalize())
        if len(candidate_monoalts) == 0:
            return None

        # create a range which covers all candidate mono-ALT vcfspecs
        merged_ref_range0 = range(
            min(x.start0 for x in candidate_monoalts),
            max(x.end0 for x in candidate_monoalts),
        )
        # make rpplist using the range with each sample
        rpplist_dict = dict()
        for sampleid, pileup in self.pileup_dict.items():
            rpplist_dict[sampleid] = readplus.get_rpplist_nonsv(
                bam=pileup.bam, 
                fasta=self.fasta, 
                chromdict=refgenome.get_default_chromdict(self.refver),
                chrom=self.chrom,
                start0=merged_ref_range0.start,
                end0=merged_ref_range0.stop,
            )
        # calculate vaf for each mono-ALT vcfspec from rpplists
        # choose if vaf is greater than cutoff in any sample
        selected_monoalt_vcfspecs = set()
        for monoalt_vcfspec in candidate_monoalts:
            if verbose:
                print()
                print(monoalt_vcfspec)

            for sampleid, rpplist in rpplist_dict.items():
                rpplist.update_alleleclass(monoalt_vcfspec)
                alleleclass_counter = collections.Counter(
                    rpp.alleleclass[monoalt_vcfspec] for rpp in rpplist
                )
                total_reads = sum(
                    v for (k, v) in alleleclass_counter.items()
                    if k is not None
                )
                var_reads = alleleclass_counter[1]
                if total_reads == 0:
                    vaf = None
                else:
                    vaf = var_reads / total_reads

                if verbose:
                    print(f'sampleid {sampleid}')
                    print(f'vaf {vaf}')
                    
                if (vaf is not None) and (vaf >= vaf_cutoff):
                    selected_monoalt_vcfspecs.add(monoalt_vcfspec)
                    break

        if len(selected_monoalt_vcfspecs) == 0:
            return None

        # sort selected vcfspecs
        selected_monoalt_vcfspecs = sorted(
            selected_monoalt_vcfspecs,
            key=(lambda x: (x.pos, x.alts)),
        )
        # merge and return
        return functools.reduce(libvcfspec.merge, selected_monoalt_vcfspecs)

    def get_result_vcfspec_recalc_hits_dual_filtering(
        self, vaf_cutoff=0.01, subseq_portion_threshold=None, 
        MQ_threshold=40, threshold_bysample=True,
        verbose=False,
    ):
        # set params
        if subseq_portion_threshold is None:
            subseq_portion_threshold = self.params['allele_portion_threshold']

        # 1st filtering
        # make candidate mono-ALT vcfspecs
        candidate_monoalts = list()
        for x in self.iter_contig_vcfspecs(
            subseq_portion_threshold=subseq_portion_threshold, 
            MQ_threshold=MQ_threshold, 
            threshold_bysample=threshold_bysample,
        ):
            if x[1] is not None:
                candidate_monoalts.append(x[1].normalize())

        if len(candidate_monoalts) == 0:
            return None

        # 2nd filtering
        # create rpplists for each component sample
        merged_ref_range0 = range(
            min(x.start0 for x in candidate_monoalts),
            max(x.end0 for x in candidate_monoalts),
        )
        rpplist_dict = dict()
        for sampleid, pileup in self.pileup_dict.items():
            rpplist_dict[sampleid] = readplus.get_rpplist_nonsv(
                bam=pileup.bam, 
                fasta=self.fasta, 
                chromdict=refgenome.get_default_chromdict(self.refver),
                chrom=self.chrom,
                start0=merged_ref_range0.start,
                end0=merged_ref_range0.stop,
            )
        # calculate vaf for each mono-ALT vcfspec from rpplists
        # choose if vaf is greater than cutoff in any sample
        selected_monoalt_vcfspecs = set()
        for monoalt_vcfspec in candidate_monoalts:
            if verbose:
                print()
                print(monoalt_vcfspec)

            for sampleid, rpplist in rpplist_dict.items():
                rpplist.update_alleleclass(monoalt_vcfspec)
                alleleclass_counter = collections.Counter(
                    rpp.alleleclass[monoalt_vcfspec] for rpp in rpplist
                )
                total_reads = sum(
                    v for (k, v) in alleleclass_counter.items()
                    if k is not None
                )
                var_reads = alleleclass_counter[1]
                if total_reads == 0:
                    vaf = None
                else:
                    vaf = var_reads / total_reads

                if verbose:
                    print(f'sampleid {sampleid}')
                    print(f'vaf {vaf}')
                    
                if (vaf is not None) and (vaf >= vaf_cutoff):
                    selected_monoalt_vcfspecs.add(monoalt_vcfspec)
                    break

        if len(selected_monoalt_vcfspecs) == 0:
            return None

        # sort selected vcfspecs
        selected_monoalt_vcfspecs = sorted(
            selected_monoalt_vcfspecs,
            key=(lambda x: (x.pos, x.alts)),
        )
        # merge and return
        return functools.reduce(libvcfspec.merge, selected_monoalt_vcfspecs)

    def get_result_vcfspec_use_rowspecgroup_hits(
        self, 
        #as_components=False, 
        #merge=True, 
        subseq_portion_threshold=None, 
        MQ_threshold=30, 
        threshold_bysample=True,
        verbose=False,
    ):
        if subseq_portion_threshold is None:
            subseq_portion_threshold = self.params['allele_portion_threshold']

#        if as_components:
#            result = list()
#            for row_id, contig_vcfspec in self.iter_contig_vcfspecs(subseq_portion_threshold=subseq_portion_threshold, MQ_threshold=MQ_threshold, threshold_bysample=threshold_bysample):
#                if contig_vcfspec is None:
#                    continue
#
#                if len(contig_vcfspec.components[0]) == 0:
#                    result.append(contig_vcfspec)
#                else:
#                    result.extend(contig_vcfspec.components[0])
        result = list(
            x[1] for x in self.iter_contig_vcfspecs(
                subseq_portion_threshold=subseq_portion_threshold, 
                MQ_threshold=MQ_threshold, 
                threshold_bysample=threshold_bysample,
                verbose=verbose,
            )
            if x[1] is not None
        )

        if len(result) == 0:
            return None
        else:
            return functools.reduce(libvcfspec.merge, result)

    # SELECT AN APPROPRIATE FUNCTION
    get_result_vcfspec = get_result_vcfspec_use_rowspecgroup_hits

    # get_result_vcfspec END #

    def show_row_spec_alignments(self, skip_zero_hits=True, show_subseqs=False):
        self._show_row_spec_alignments_helper(
            self.row_spec_groups, 
            lambda x: self.pileup_dict[x[0]].row_specs.__getitem__(x[1]),
            lambda x: self.pileup_dict[x[0]]._superseq_alignments.__getitem__(x[1]),
            skip_zero_hits, 
            show_subseqs,
        )



import textwrap

import pysam

import handygenome.refgenome as refgenome
import handygenome.workflow as workflow
import handygenome.workflow.toolsetup as toolsetup
import handygenome.variant.svcaller_parser as svcaller_parser
import handygenome.variant.varianthandler as varianthandler
import handygenome.vcfeditor.headerhandler as headerhandler
import handygenome.vcfeditor.indexing as indexing
import handygenome.variant.vcfspec as libvcfspec


def argument_parser(cmdargs):
    def sanity_check(args):
        pass

    parser_dict = workflow.init_parser(
        description=textwrap.dedent(f"""\
            - Split multiallelic lines.
            - Change SV caller records into breakends form.
            - Normalize non-SV records into the leftmost form.
            - Normalize SV records into the bnd1-advanced form."""
        )
    )

    workflow.add_infile_arg(parser_dict['required'], required=True)
    workflow.add_outfile_arg(parser_dict['required'], required=True, must_not_exist='ask')
    #workflow.add_fasta_arg(parser_dict['required'], required=True)
    workflow.add_refver_arg(parser_dict['required'], required=True, choices='all')
    workflow.add_outfmt_arg(parser_dict['optional'], required=False)

    workflow.add_logging_args(parser_dict)

    workflow.add_index_arg(parser_dict)

    args = parser_dict['main'].parse_args(cmdargs)
    sanity_check(args)

    return args


def process_nonsv(vr, out_vr_list, nonsv_set, logger, refver):
    """
    Normalize the input vr(the leftmost form).
    Check if the normalized vcfspec is duplicated, in which case the vr is
        discarded.
    """

    vcfspec_original = libvcfspec.Vcfspec.from_vr(vr, refver=refver)
    vcfspec_left = vcfspec_original.leftmost()
    ID = vcfspec_left.get_id()

    if ID in nonsv_set:
        logger.info(f'Input variant record is discarded since vcfspec is duplicated:\n{vr}')
    else:
        nonsv_set.add(ID)
        varianthandler.apply_vcfspec(vr, vcfspec_left)
        vr.id = ID
        out_vr_list.append(vr)


def process_sv(vr, out_vr_list, bnds_set, fasta, chromdict, logger):
    """
    From a SV vr, generate two variant records representing bnd1 and bnd2.
    The breakends are in pos1adv form.
    Check if the bndspec is duplicated, in which case the vr is discarded.
    """

    bnds = svcaller_parser.get_bnds_from_caller_vr(vr, fasta, chromdict) 

    ##
    ##

    if bnds is None:
        raise Exception(f'Error while parsing this SV variant record:\n{vr}')

    if bnds in bnds_set:
        logger.info(f'Input variant record is discarded since breakends '
                    f'spec is duplicated:\n{vr}')
        return
    else:
        bnds_set.add(bnds)
        vcfspec_bnd1 = bnds.get_vcfspec_bnd1()
        vcfspec_bnd2 = bnds.get_vcfspec_bnd2()
        ID_bnd1 = bnds.get_id_bnd1()
        ID_bnd2 = bnds.get_id_bnd2()

        varianthandler.apply_vcfspec(vr, vcfspec_bnd1)
        vr.id = ID_bnd1
        vr.info['MATEID'] = ID_bnd2

        vr_pos2 = vr.header.new_record()
        varianthandler.apply_vcfspec(vr_pos2, vcfspec_bnd2)
        vr_pos2.id = ID_bnd2
        vr_pos2.info['MATEID'] = ID_bnd1

        out_vr_list.append(vr)
        out_vr_list.append(vr_pos2)


def process_monoallelic_vr(vr, out_vr_list, nonsv_set, bnds_set, fasta, chromdict, logger, refver):
    # skip delly <INS> records
    if svcaller_parser.check_vr_format_delly(vr):
        if vr.alts[0] == '<INS>':
            return

    if varianthandler.check_SV(vr):
        process_sv(vr, out_vr_list, bnds_set, fasta, chromdict, logger)
    else:
        process_nonsv(vr, out_vr_list, nonsv_set, logger, refver)


def main(cmdargs):
    args = argument_parser(cmdargs)
    logger = toolsetup.setup_logger(args)

    logger.info('BEGINNING')

    fasta = refgenome.get_default_fasta(args.refver)
    chromdict = refgenome.get_default_chromdict(args.refver)
    refver = args.refver

    nonsv_set = set()
    bnds_set = set()
    out_vr_list = list()
    
    with pysam.VariantFile(args.infile_path) as in_vcf:
        #headerhandler.addmeta_MATEID(in_vcf.header)
        #headerhandler.addmeta_END(in_vcf.header)
        with pysam.VariantFile(args.outfile_path, mode=args.mode_pysam, header=in_vcf.header) as out_vcf:
            if vr.contig not in chromdict.contigs:
                continue

            for vr in in_vcf.fetch():
                if len(vr.alts) == 1:
                    process_monoallelic_vr(vr, out_vr_list, nonsv_set, bnds_set, fasta, chromdict, logger, refver)
                else:
                    for alt in vr.alts:
                        new_vr = vr.copy()
                        new_vr.alts = [alt]
                        process_monoallelic_vr(new_vr, out_vr_list, nonsv_set, bnds_set, fasta, chromdict, logger, refver)

            out_vr_list.sort(key=varianthandler.get_vr_sortkey(chromdict))

            for vr in out_vr_list:
                out_vcf.write(vr)

    if not args.donot_index:
        indexing.index_vcf(args.outfile_path)

    logger.info('ALL SUCCESSFULLY FINISHED')




import os

import pysam

import handygenome
import handygenome.refgenome as refgenome


def get_dbsnp_path(refver):
    standard = refgenome.REFVERINFO.standardize(refver)
    refveritem = refgenome.REFVERINFO.access_refveritem(standard)
    if refveritem['dbsnp_path'] is None:
        return os.path.join(handygenome.DIRS['data'], 'popfreq', f'dbSNP_{standard}.vcf.gz')
    else:
        return refveritem['dbsnp_path']


REPEATS_PATH = refgenome.RefverDict({
    'GRCh37': '/home/users/pjh/References/ucsc_data_files/modified_files/grch37/repeatmasker.bed.gz',
    'GRCh38': '/home/users/pjh/References/ucsc_data_files/modified_files/grch38/repeatmasker.bed.gz',
})

GENESET_PATH = refgenome.RefverDict({
    'GRCh37': '/home/users/pjh/References/ensembl_data_files/modified_files/grch37/geneset_gff3_sorted.gz',
    'GRCh38': '/home/users/pjh/References/ensembl_data_files/modified_files/grch38/geneset_gff3_sorted.gz',
})

REGULATORY_PATH = refgenome.RefverDict({
    'GRCh37': '/home/users/pjh/References/ensembl_data_files/modified_files/grch37/ensembl_regulatory_sorted.grch37.gff.gz',
    'GRCh38': '/home/users/pjh/References/ensembl_data_files/modified_files/grch38/ensembl_regulatory_sorted.grch38.gff.gz',
})

DBSNP_PATHS = refgenome.RefverDict({
    'GRCh37': f'{handygenome.DIRS["data"]}/popfreq/dbSNP_b155_GRCh37.p13.vcf.gz',
    'GRCh38': f'{handygenome.DIRS["data"]}/popfreq/dbSNP_b156_GRCh38.p13.vcf.gz',
})

COSMIC_PATHS = refgenome.RefverDict({
    'GRCh37': f'{handygenome.DIRS["data"]}/cosmic/v96/grch37/nonSV.vcf.gz',
    'GRCh38': None,
})

###

TABIXFILES_REPEATS = refgenome.RefverDict({
    'GRCh37': pysam.TabixFile(REPEATS_PATH['GRCh37'], parser=pysam.asBed()),
    'GRCh38': pysam.TabixFile(REPEATS_PATH['GRCh38'], parser=pysam.asBed()),
})

TABIXFILES_GENESET = refgenome.RefverDict({
    'GRCh37': pysam.TabixFile(GENESET_PATH['GRCh37'], parser=pysam.asGTF()),
    'GRCh38': pysam.TabixFile(GENESET_PATH['GRCh38'], parser=pysam.asGTF()),
})

TABIXFILES_REGULATORY = refgenome.RefverDict({
    'GRCh37': pysam.TabixFile(REGULATORY_PATH['GRCh37'], parser=pysam.asGTF()),
    'GRCh38': pysam.TabixFile(REGULATORY_PATH['GRCh38'], parser=pysam.asGTF()),
})

VCFS_DBSNP = refgenome.RefverDict({
    'GRCh37': pysam.VariantFile(DBSNP_PATHS['GRCh37']),
    'GRCh38': pysam.VariantFile(DBSNP_PATHS['GRCh38']),
})

VCFS_COSMIC = refgenome.RefverDict({
    'GRCh37': pysam.VariantFile(COSMIC_PATHS['GRCh37']),
    'GRCh38': None,
})



import re
import os
import socket

import numpy as np

import handygenome.sv.breakends as breakends
import handygenome.variant.vcfspec as libvcfspec


class IGVHandle:
    def __init__(self, port, pathsub=None):
        self.port = port
        self.pathsub = pathsub

    def cmd(self, msg):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.connect(('localhost', self.port))
            s.send(msg.encode('utf-8'))
            s.shutdown(socket.SHUT_WR)
            data = s.recv(1024).decode()
            print(data, end='')

    def new(self):
        self.cmd('new')
        
    def goto(self, loci, width=200):
        assert isinstance(loci, (list, tuple)), f'"loci" argument must be a list'

        cmd_src = list()
        for locus in loci:
            if isinstance(locus, tuple):
                cmd_src.append(f'{locus[0]}:{locus[1] - width}-{locus[2] + width}')
            elif isinstance(locus, str):
                locus_sp = locus.split(':')
                chrom = locus_sp[0]
                start1, end1 = locus_sp[1].split('-')
                start1 = int(start1)
                end1 = int(end1)
                cmd_src.append(f'{chrom}:{start1 - width}-{end1 + width}')
            elif isinstance(locus, libvcfspec.Vcfspec):
                chrom = locus.chrom
                if locus.get_mutation_type(alt_index=0) == 'del':
                    start1 = locus.pos + 1 - width
                    end1 = locus.pos + 1 + width
                else:
                    start1 = locus.pos - width
                    end1 = locus.pos + width

                cmd_src.append(f'{chrom}:{start1}-{end1}')
            elif isinstance(locus, breakends.Breakends):
                chrom1 = locus.chrom_bnd1
                start1 = locus.pos_bnd1 - width
                end1 = locus.pos_bnd1 + width
                cmd_src.append(f'{chrom1}:{start1}-{end1}')

                chrom2 = locus.chrom_bnd2
                start2 = locus.pos_bnd2 - width
                end2 = locus.pos_bnd2 + width
                cmd_src.append(f'{chrom2}:{start2}-{end2}')
            else:
                raise Exception(
                    f'Entries of "loci" must be a str, tuple, Vcfspec, or Breakends.'
                )

        self.cmd(f'goto {" ".join(cmd_src)}')

    def load(self, filepaths):
        # make into absolute paths
        filepaths = [os.path.abspath(x) for x in np.atleast_1d(filepaths)]
        # do substition
        if self.pathsub is not None:
            filepaths = [re.sub(self.pathsub[0], self.pathsub[1], x) for x in filepaths]

        for path in filepaths:
            self.cmd(f'load {path}')
    
    def zoomin(self, n=1):
        for _ in range(n):
            self.cmd('zoomin')

    def zoomout(self, n=1):
        for _ in range(n):
            self.cmd('zoomout')

    def viewaspairs(self):
        self.cmd('viewaspairs')

    def viewaspairs_off(self):
        self.cmd('viewaspairs false')

    def snapshot(self, path, maxPanelHeight=1000):
        basename = os.path.basename(path)
        dirname = os.path.dirname(path)

        if not os.path.exists(dirname):
            os.makedirs(dirname)
        self.cmd(f'snapshotDirectory {dirname}')

        self.cmd(f'maxPanelHeight {maxPanelHeight}')
        self.cmd(f'snapshot {basename}')


        



import os
import re
import tempfile
import subprocess

import numpy as np
import pandas as pd
import matplotlib as mpl
import matplotlib.pyplot as plt
import scipy
import pyranges as pr

import handygenome
import handygenome.common as common
import handygenome.cnv.read_cnvfile as read_cnvfile
import handygenome.pyranges_helper as pyranges_helper


CONDA_WRAPPER = os.path.join(handygenome.DIRS['scripts'], 'conda_wrapper.sh')
RUN_EXTRACT = os.path.join(handygenome.DIRS['R'], 'sequenza_extract.R')
RUN_FIT = os.path.join(handygenome.DIRS['R'], 'sequenza_fit.R')
WRITE_EXTRACT_TSV = os.path.join(handygenome.DIRS['R'], 'sequenza_extract_to_tsv.R')
WRITE_FIT_TSV = os.path.join(handygenome.DIRS['R'], 'sequenza_fit_to_tsv.R')
MAKE_SEGMENTS = os.path.join(handygenome.DIRS['R'], 'sequenza_results_segment.R')

CONDA_WRAPPER_ARGS = [
    CONDA_WRAPPER,  
    '--condadir', '/home/users/pjh/tools/miniconda/220718_for_sequenza/miniconda3/',
    '--envname', 'sequenza',
    'Rscript',
]

RSCRIPT_PATH = '/home/users/pjh/tools/miniconda/220718_for_sequenza/miniconda3/envs/sequenza/bin/Rscript'


def make_segments_as_df(extractfile_path, cellularity, ploidy, is_female):
    fd, tsv_path = tempfile.mkstemp(
        prefix='tmpfile_sequenza_segments_', suffix='.tsv', dir=os.getcwd(),
    )
    os.close(fd)
    make_segments(extractfile_path, cellularity, ploidy, is_female, tsv_path)
    df = read_sequenza_segment(tsv_path)
    os.remove(tsv_path)
    return df


def make_segments(extractfile_path, cellularity, ploidy, is_female, outfile_path):
    args = list()
    #args.extend(CONDA_WRAPPER_ARGS)
    args.append(RSCRIPT_PATH)
    args.extend(
        [
            MAKE_SEGMENTS,
            '--extfile', extractfile_path,
            '--outfile', outfile_path,
            '--cellularity', str(cellularity),
            '--ploidy', str(ploidy),
            '--gender', ('F' if is_female else 'M'),
        ]
    )
    p = subprocess.run(args, capture_output=True, text=True, check=True)


def read_sequenza_segment(segment_path, as_gr=False):
    df = pd.read_table(
        segment_path, 
        names=(
            'Chromosome', 'Start', 'End',
            'Bf', 'N.BAF', 'sd.BAF',
            'depth.ratio', 'N.ratio', 'sd.ratio',
            'CNt', 'A', 'B', 'LPP',
        ),
        skiprows=1,
    )
    df.loc[df['CNt'] == 0, 'A'] = 0
    df.loc[df['CNt'] == 0, 'B'] = 0

    if as_gr:
        return pr.PyRanges(df=df)
    else:
        return df


def load_fit(fitfile_path):
    fd, tsv_path = tempfile.mkstemp(
        prefix='tmpfile_sequenza_fit_result_', suffix='.tsv', dir=os.getcwd(),
    )
    os.close(fd)

    args = list()
    #args.extend(CONDA_WRAPPER_ARGS)
    args.append(RSCRIPT_PATH)
    args.extend(
        [
            WRITE_FIT_TSV,
            '--fitfile', fitfile_path,
            '--outfile', tsv_path,
        ]
    )
    p = subprocess.run(args, capture_output=True, text=True, check=True)

    df = pd.read_table(tsv_path, sep='\t', index_col=0)
    df.columns = df.columns.astype('float')
    df.columns.name = 'cellularity'
    df.index.name = 'ploidy'

    os.remove(tsv_path)

    return df


def load_extract(extractfile_path, chromdict, as_gr=True):
    # make temp file
    fd, tsv_path = tempfile.mkstemp(
        prefix='tmpfile_sequenza_extract_result_', suffix='.tsv', dir=os.getcwd(),
    )
    os.close(fd)

    # run R script
    args = list()
    #args.extend(CONDA_WRAPPER_ARGS)
    args.append(RSCRIPT_PATH)
    args.extend(
        [
            WRITE_EXTRACT_TSV,
            '--extractfile', extractfile_path,
            '--outfile', tsv_path,
        ]
    )
    p = subprocess.run(args, capture_output=True, text=True, check=False)
    if p.returncode != 0:
        print(f'stdout: {p.stdout}')
        print(f'stderr: {p.stderr}')
        p.check_returncode()

    # load dataframe
    df = pd.read_table(tsv_path, sep='\t', header=0, dtype={'chrom': 'string'})
    df.rename(columns={'start': 'Start', 'end': 'End', 'chrom': 'Chromosome'}, inplace=True)
    new_columns = ['Chromosome', 'Start', 'End'] + df.columns.drop(['Chromosome', 'Start', 'End']).to_list()
    df = df.loc[:, new_columns]
    df = pyranges_helper.sort_df_by_coord(df, chromdict)

    os.remove(tsv_path)

    if as_gr:
        return pr.PyRanges(df, int64=False)
    else:
        return df


def get_1d_peaks(row_iterator):
    result = list()
    for row in row_iterator:
        row_peaks = list()
        peaks_result, _ = scipy.signal.find_peaks(row)
        diff = np.diff(row)
        for within_row_idx in peaks_result:
            row_peaks.append(within_row_idx)
            # leftward
            current_idx = within_row_idx
            while True:
                current_idx -= 1
                if diff[current_idx] != 0:
                    break
                else:
                    row_peaks.append(current_idx)

                if current_idx == 0:
                    break
            # rightward
            current_idx = within_row_idx - 1
            while True:
                current_idx += 1
                if diff[current_idx] != 0:
                    break
                else:
                    row_peaks.append(current_idx + 1)

                if current_idx == len(diff) - 1:
                    break

        row_peaks.sort()
        result.append(row_peaks)
        
    return result


def get_fitresult_peaks(fitresult):
    row_peaks = get_1d_peaks((x[1] for x in fitresult.iterrows()))
    col_peaks = get_1d_peaks((x[1] for x in fitresult.items()))

    row_peaks_tuples = list()
    for row_idx, peaks in enumerate(row_peaks):
        for col_idx in peaks:
            row_peaks_tuples.append((row_idx, col_idx))

    col_peaks_tuples = list()
    for col_idx, peaks in enumerate(col_peaks):
        for row_idx in peaks:
            col_peaks_tuples.append((row_idx, col_idx))

    result = set(row_peaks_tuples).intersection(set(col_peaks_tuples))
    result = [
        {'ploidy': fitresult.index[x], 'cellularity': fitresult.columns[y], 'lpp': fitresult.iloc[x, y]}
        for (x, y) in result
    ]
    result.sort(key=(lambda x: x['lpp']), reverse=True)
    
    return result





import pprint
import os
import re
import gzip
import shutil
import itertools

import scipy.stats
import numpy as np

import handygenome.deco as deco


PAT_INT = re.compile('-?[0-9]+')
PAT_FLOAT = re.compile('(-?[0-9]+\.[0-9]+)|(-?[0-9]+(\.[0-9]+)?e-?[0-9]+)')


# custom class base

def repr_base(obj, keylist, comma_sep_int=False):
    string_list = list()
    for key in keylist:
        val = getattr(obj, key)
        if isinstance(val, int) and comma_sep_int:
            string_list.append(f'{key}={val:,}')
        else:
            string_list.append(f'{key}={repr(val)}')

    return ', '.join(string_list)


# colors

COLORS = {
    'red':     '\033[38;5;196m',
    'magenta': '\033[38;5;201m',
    'pink':    '\033[38;5;213m',
    'orange':  '\033[38;5;9m',
    'yellow':  '\033[38;5;214m',
    'gold':    '\033[38;5;11m',
    'green':   '\033[38;5;40m',
    'blue':    '\033[38;5;33m',
    'cyan':    '\033[38;5;14m',
    'purple':  '\033[38;5;93m',
    'gray':    '\033[38;5;8m',
    'white':   '\033[38;5;15m',
    'end':     '\033[0m',
    }


class ColorsQQ:
    mine="\033[48;5;6m"
    busy="\033[48;5;244m"
    free="\033[48;5;238m"
    end="\033[0m"
    nor="\033[48;5;160m"
    nor2="\033[48;5;52m"
    danger1="\033[38;5;208m"
    danger2="\033[38;5;196m"
    darkgray="\033[38;5;240m"


def visualize_colors():
    for i in range(256):
        print(f'{i:<3d} \033[38;5;{i}m\\033[38;5;{i}m\033[0m')


def cpformat(obj, **kwargs):
    result = pprint.pformat(obj, **kwargs)
    result = re.sub('(True)', COLORS['green'] + '\\1' + COLORS['end'], result)
    result = re.sub('(False)', COLORS['red'] + '\\1' + COLORS['end'], result)
    result = re.sub('(None)', COLORS['purple'] + '\\1' + COLORS['end'], result)
    return result


def cpprint(obj):
    print(cpformat(obj))


# string manipulators

def str_to_nonstr(val):
    #assert isinstance(val, str)

    if val.lower() in ('none', 'null'):
        return None
    elif val.lower() == 'nan':
        return np.nan
    elif val.lower() == 'true':
        return True
    elif val.lower() == 'false':
        return False
    elif PAT_INT.fullmatch(val) is not None:
        return int(val)
    elif PAT_FLOAT.fullmatch(val) is not None:
        return float(val)
    else:
        return val


def nonstr_to_str(val):
    assert not isinstance(val, str)

    if val is None:
        return None
    else:
        return str(val)


def zrange(n):
    assert n > 0
    width = len(str(n - 1))
    for idx in range(n):
        zidx = str(idx).zfill(width)
        yield zidx


def zenumerate(iterable):
    iterable_tup = tuple(iterable)
    length = len(iterable_tup)
    for zidx, item in zip(zrange(length), iterable_tup):
        yield zidx, item


def zidx_to_idx(zidx):
    if re.match('^0+$', zidx):
        return 0
    else:
        return int(re.sub(r'^0*', '', zidx))


def get_padded_indices(n):
    """Begins with 0"""
    width = len(str(n-1))
    for idx in range(n):
        yield str(idx).zfill(width)
    #result = [str(idx).zfill(width) for idx in range(n)]
    #return result


def shorten_int(numlist, n_after_dot=3):
    mapping = {0: '', 1: 'K', 2: 'M', 3: 'G'}

    log = np.log10(numlist)
    assert (log < 12).all(), f'Numbers greater than or equal to 10^12 are not allowed.'

    #qs = [int(y) for y in (log / 3)]
    qs = np.floor(log / 3)

    suffixes = [mapping[x] for x in qs]
    #new_numlist = numlist / (10 ** (np.array(qs) * 3))
    new_numlist = numlist / (10 ** (qs * 3))
    formatter = f'.{n_after_dot}f'
    result = [
        (
            f'{int(x)}' 
            if y == '' else 
            f'{{x:{formatter}}} {{y}}'.format(x=x, y=y)
        )
        for x, y in zip(new_numlist, suffixes)
    ]
    return result


"""
http://code.activestate.com/recipes/578019
https://psutil.readthedocs.io/en/latest/#bytes-conversion
"""
_bytes2human_symbols = ('K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y')
_bytes2human_symbols_rev = tuple(reversed(_bytes2human_symbols))
_bytes2human_prefix = {
    s: 1 << (i + 1) * 10
    for i, s in enumerate(_bytes2human_symbols)
}

def bytes2human(n):
    for s in _bytes2human_symbols_rev:
        if abs(n) >= _bytes2human_prefix[s]:
            value = float(n) / _bytes2human_prefix[s]
            return '%.1f%s' % (value, s)
    return "%sB" % n


# listdir

def listdir(path):
    return sorted(os.path.join(path, x) for x in os.listdir(path))


# file handling

def unzip(src, dest, rm_src=False):
    with gzip.open(src, 'rb') as infile:
        with open(dest, 'wb') as outfile:
            shutil.copyfileobj(infile, outfile)      
    if rm_src:
        os.remove(src)


@deco.get_deco_arg_choices({'mode': ('r', 'w', 'a')})
def openfile(fname, mode='r'):
    mode = mode + 't'

    if fname.endswith('.gz'):
        return gzip.open(fname, mode)
    else:
        return open(fname, mode)


def fileiter(path, sep='\t', remove_leading_hashes=True, skip_double_hashes=True):
    infile = openfile(path, 'r')

    while True:
        line = next(infile)
        if skip_double_hashes:
            if line.startswith('##'):
                continue
            else:
                break
        else:
            break
        
    headerline = line
    if remove_leading_hashes:
        headerline = re.sub('^#*', '', headerline)
    header = get_linesp(headerline, sep=sep)

    for line in infile:
        linesp = get_linesp(line, sep=sep)
        if len(linesp) != len(header):
            raise Exception(
                f'Field numbers of the header line and the current '
                f'line are different:\n'
                f'header: {header}\n'
                f'line: {linesp}')
        linedict = dict(zip(header, linesp))
        yield linedict

    infile.close()


def rm_newline(line):
    return re.sub('(\r)?\n$', '', line)


def rm_newline_byte(byteline):
    return re.sub(b'(\r)?\n$', b'', byteline)


def get_linesp(line, sep='\t'):
    return rm_newline(line).split(sep)


def get_linesp_byte(byteline, sep=b'\t'):
    return rm_newline_byte(byteline).split(sep)


# printwidth

def printwidth_get_width_list(df):
    '''
    df: [
    [line1_field1, line1_field2, ... ],
    [line2_field1, line2_field2, ... ],
    ...,
    ]
    '''
    width_list = list()
    for i in range(len(df[0])):
        width_list.append(list())

    for line in df:
        for idx, field in enumerate(line):
            width_list[idx].append(len(field))

    for idx, e in enumerate(width_list):
        width_list[idx] = max(e)

    return width_list


def printwidth_print_line(line, width_list, margin, target):
    printresult = ''
    for idx, e in enumerate(line):
        printresult += f'{e:>{width_list[idx] + margin}s}'
    if target == 'out':
        print(printresult, flush = True)
    elif target == 'err':
        print(printresult, flush = True, file = sys.stderr)


def printwidth(df, margin = 2, target = 'out'):
    for line in df:
        for idx, e in enumerate(line):
            line[idx] = str(e)

    width_list = printwidth_get_width_list(df)
    for line in df:
        printwidth_print_line(line, width_list, margin, target)


# memory usage check
def get_rss(mode='total', unit='b'):
    assert mode in ('self', 'children', 'total')
    assert unit in ('b', 'k', 'm', 'g')

    proc = psutil.Process()
    children = proc.children(recursive=True)
    rss_self = proc.memory_info().rss  # in bytes
    rss_children = sum(p.memory_info().rss for p in children)  # in bytes

    if mode == 'self':
        rss = rss_self
    elif mode == 'children':
        rss = rss_children
    elif mode == 'total':
        rss = rss_self + rss_children

    if unit == 'b':
        rss = rss
    elif unit == 'k':
        rss = rss / 1024
    elif unit == 'm':
        rss = rss / 1024**2
    elif unit == 'g':
        rss = rss / 1024**3

    return rss


# multi-min, multi-max

def _multi_selector_base(sequence, comparing_func, key=None, with_target_val=False):
    sequence = list(sequence)
    if key is None:
        values = sequence
    else:
        values = [key(x) for x in sequence]
    target_val = comparing_func(values)
    is_target_val = [x == target_val for x in values]

    if with_target_val:
        return list(itertools.compress(sequence, is_target_val)), target_val
    else:
        return list(itertools.compress(sequence, is_target_val))


def multi_min(sequence, key=None, with_target_val=False):
    return _multi_selector_base(sequence, min, key=key, with_target_val=with_target_val)


def multi_max(sequence, key=None, with_target_val=False):
    return _multi_selector_base(sequence, max, key=key, with_target_val=with_target_val)



# itertools plus utilities

def pairwise(iterable):
    iterable = iter(iterable)
    try:
        x0 = next(iterable)
        x1 = next(iterable)
    except StopIteration:
        return

    while True:
        yield (x0, x1)
        try:
            x2 = next(iterable)
        except StopIteration:
            return
        x0 = x1
        x1 = x2
        continue


def chunk_iter(iterable, n):
    args = [iter(iterable)] * n
    for subiter in itertools.zip_longest(*args, fillvalue=None):
        yield tuple(x for x in subiter if x is not None)


# from Itertools Recipes; similar as chunk_iter
def grouper_Itertools_Recipes(iterable, n, *, incomplete='fill', fillvalue=None):
    "Collect data into non-overlapping fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3, fillvalue='x') --> ABC DEF Gxx
    # grouper('ABCDEFG', 3, incomplete='strict') --> ABC DEF ValueError
    # grouper('ABCDEFG', 3, incomplete='ignore') --> ABC DEF
    args = [iter(iterable)] * n
    if incomplete == 'fill':
        return itertools.zip_longest(*args, fillvalue=fillvalue)
    if incomplete == 'strict':
        return zip(*args, strict=True)
    if incomplete == 'ignore':
        return zip(*args)
    else:
        raise ValueError('Expected fill, strict, or ignore')



# array handlers & mathematical ones

def round_sig(num, n):
    """round 'num' with 'n' significant digits"""

    assert n > 0
    from math import log10, floor
    return round(num, -floor(log10(abs(num))) + (n-1))


def get_split_nums(total_length, num):
    arrsp = np.array_split(np.zeros(total_length), num)
    return np.fromiter(
        (x.shape[0] for x in arrsp if x.shape[0] != 0), 
        dtype=int,
    )


def array_grouper(arr, omit_values=False):
    assert arr.ndim in (1, 2)

    diff = np.empty(arr.shape[0], dtype=bool)
    diff[0] = True
    if arr.ndim == 1:
        diff[1:] = np.diff(arr)
    elif arr.ndim == 2:
        diff[1:] = np.diff(arr, axis=0).any(axis=1)

    indexes = np.nonzero(diff)[0]

    if omit_values:
        values = None
    else:
        values = arr[indexes]

    counts = np.empty(indexes.shape, dtype=int)
    counts[:-1] = np.diff(indexes)
    counts[-1] = arr.shape[0] - indexes[-1]

    groupkey = np.repeat(np.arange(len(counts)), counts)
        
    return values, counts, groupkey
                

def get_indexes_of_array(values, ordered_keys):
    """Used for handling pandas.cut function output, in cnv.misc module"""
    result = np.zeros(len(values))
    for idx, key in enumerate(ordered_keys):
        result[np.where(values == key)[0]] = idx

    return result


def nanaverage(values, weights):
    #assert isinstance(values, np.ndarray)
    #assert isinstance(weights, np.ndarray)

    values = np.array(values)
    weights = np.array(weights)

    selector = ~np.isnan(values)
    new_values = values[selector]
    new_weights = weights[selector]
    return np.average(new_values, weights=new_weights)


def get_ranks(arr):
    ranks = scipy.stats.rankdata(data, method='max')
    return ranks / len(ranks)


def get_mode(data, xs=np.arange(0, 0.51, 0.01)):
    kernel = scipy.stats.gaussian_kde(data)
    ys = kernel(xs)
    return xs[np.argmax(ys)]


def bernoulli_iterator(iterable, p, block_size=int(1e5)):
    """Args:
        p: probability of selection
    """
    for x, rv in zip(iter(iterable), bernoulli_rv_generator(p, block_size=block_size)):
        if rv:  # True if 1
            yield x


def bernoulli_rv_generator(p, block_size=int(1e5)):
    while True:
        rvs = scipy.stats.bernoulli.rvs(p=p, loc=0, size=block_size)
        for x in rvs:
            yield x


def mean_mad(values):
    values = np.array(values)
    return np.mean(np.abs(values - np.mean(values)))


def median_mad(values):
    values = np.array(values)
    return np.median(np.abs(values - np.median(values)))


def get_diffmean(values, weights=None):
    assert values.ndim == 1

    if weights is None:
        weights = np.ones_like(values)
    else:
        assert weights.shape == values.shape

    indexes = np.triu_indices(values.shape[0], k=1)
    diffs = values[indexes[0]] - values[indexes[1]]
    diff_weights = weights[indexes[0]] + weights[indexes[1]]

    return np.average(np.abs(diffs), weights=diff_weights)


##############################
# Genomic coordinate sorting #
##############################

def coord_sortkey(chrom, pos, chromdict): 
    """
    Args:
        pos: 1-based
        chromdict: ChromDict class instance
    """

    return (chromdict.contigs.index(chrom), pos)


#def get_vcfspec_sortkey(chromdict):
#    def sortkey(vcfspec):
#        return (chromdict.contigs.index(vcfspec.chrom), vcfspec.pos, vcfspec.ref) + vcfspec.alts
#        #return coord_sortkey(vcfspec.chrom, vcfspec.pos, chromdict)
#
#    return sortkey


def compare_coords(chrom1, pos1, chrom2, pos2, chromdict):
    """
    Returns:
        0: equal; -1: chrom1/pos1 comes first; 1: chrom2/pos2 comes first
    """

    if chrom1 == chrom2 and pos1 == pos2:
        return 0
    else:
        if (coord_sortkey(chrom1, pos1, chromdict) 
            < coord_sortkey(chrom2, pos2, chromdict)):
            return -1
        else:
            return 1



##########################################
# Check if file is plain text or gzipped #
##########################################

def check_file_is_plain(fname):
    with open(fname, 'r') as f:
        try:
            _ = f.read(1)
        except UnicodeDecodeError:
            return False
        else:
            return True


def check_file_is_gzipped(fname):
    with gzip.open(fname, 'rb') as f:
        try:
            _ = f.read(1)
        except gzip.BadGzipFile as exc:
            if str(exc).startswith('Not a gzipped file'):
                return False
            else:
                raise
        else:
            return True
    

##################################
# temporary file path generation #
##################################

def get_tmpfile_path(prefix=None, suffix=None, dir=None, where=None, delete=False, is_dir=False):
    """Args:
        where: alias for dir
    """
    # alias handling
    assert not ((dir is not None) and (where is not None))
    if where is not None:
        dir = where
    # main
    if dir is None:
        dir = os.getcwd()

    if delete:
        fd, path = tempfile.mkstemp(prefix=prefix, suffix=suffix, dir=dir)
        os.close(fd)
        os.remove(path)
    else:
        if is_dir:
            path = tempfile.mkdtemp(prefix=prefix, suffix=suffix, dir=dir)
        else:
            fd, path = tempfile.mkstemp(prefix=prefix, suffix=suffix, dir=dir)
            os.close(fd)

    return path



#!/usr/bin/env python3

# This source code file is a part of SigProfilerSimulator
 
#Author: Erik Bergstrom

#Contact: ebergstr@eng.ucsd.edu

from __future__ import print_function
import sys
import os
import re
import argparse

revcompl = lambda x: ''.join([{'A':'T','C':'G','G':'C','T':'A','N':'N'}[B] for B in x][::-1])
revbias = lambda x: ''.join([{'0':'0', '3':'3', '1':'2','2':'1','U':'T','T':'U','B':'B','N':'N', 'Q':'Q'}[B] for B in x][::-1])


def context_distribution (context_input, output_file, chromosome_path, chromosomes, tsb_ref, genome):
	'''
	Creates a csv file for the distribution of nucleotides given a specific context. 
	This csv file needs to be created before simulating mutationalsigantures for the 
	given context.

	Requires: 
		Chromosomes saved in individual text files ('X.txt','Y.txt', '1.txt', etc)
		Transcriptional data saved in binary files for each chromosome. These files
			can be created using the script: "save_tsb_192.py"

	Parameters:
			  context_input  -> simulation context of interest (ex: 96, 192, 1536, 3072, DINUC, INDEL)
				output_file  -> file where the distribution for the given nucleotide context is saved (csv file)
			chromosome_path  -> path to the reference chromosomes
				chromosomes  -> list of chromosomes for the species of interest

	Returns:
		None

	Outputs:
		CSV file with each chromosome represented as a column and reach row 
		represented as a nucleotide. Under each chromosome for a given nucleotide
		is the proportion of the total length of that chromosome associated with that 
		nucleotide. 
	'''


	dinuc_types = ['AA', 'AC', 'AG', 'AT', 'CA', 'CC', 'CG', 'GA', 'GC', 'TA']
	dinuc_tsb = ['AA', 'AG', 'CC', 'GA',]
	dinuc_non_tsb = ['Q:AC', 'Q:AT', 'Q:CA', 'Q:CG', 'Q:GC', 'Q:TA']
	tsb_bias = ['T','U','B','N']

	# Set the context parameter based upon the user input
	if context_input == '96' or context_input == '192' or context_input == "384":
		context = 3
	elif context_input == '1536' or context_input == '3072' or context_input == '6144':
		context = 5
	elif context_input == 'DINUC' or context_input == 'DBS186' or context_input == "DBS":
		context = 2
	elif context_input == '6' or context_input == '24':
		context = 1
	else:
		print('Not a valid context')
		sys.exit()

	count = 0 
	probs = {}
	chromosome_lengths = []


	# Populate the dictionary if desired context is DINUC
	if context_input == 'DBS' or context_input == 'DINUC': 
		for dinuc in dinuc_types:
			probs[dinuc] = {}

	if context_input == 'DBS186':
		for dinuc in dinuc_tsb:
			for bias in tsb_bias:
				dinuc_t = bias + ":" + dinuc
				probs[dinuc_t] = {}
		for dinuc in dinuc_non_tsb:
			probs[dinuc] = {}


	# Iterate through each chromosome and open the associated file
	for chrom in chromosomes:
		with open (chromosome_path + chrom + ".txt", "rb") as f:

			chromosome = f.read().strip()
			chromosome_lengths.append(len(chromosome))
			print(chrom, len(chromosome))
			# If desired context is for TSB, open the transcriptional file, too
			# if context_input == '192' or context_input == '3072':
			# 	with open (chromosome_TSB_path + chrom + "_192.txt", 'rb') as tsb:
			# 		chromosome_tsb = tsb.read()

			# Iterate through the chromosome base by base
			for i in range (0, (len(chromosome)-context), 1):
				nuc = ''
				for l in range (i, i+context, 1):
					nuc += tsb_ref[chromosome[l]][1]
				base = nuc[int(context/2)]

				count += 1
				if count == 1000000:
					print(i)
					count = 0
				# Skip the base if unknown 
				if "N" in nuc:
					pass

				else:
					if context_input != "DINUC" and context_input != "DBS186":
						# Only save the pyrimidine context (canonical)
						if base == 'A' or base == 'G':
							nuc = revcompl(nuc)


						# Adjust the nucleotide representaiton if TSB is desired
						if context_input == '192' or context_input == '3072' or context_input == '384' or context_input == '6144' or context_input == '24':
							bias = tsb_ref[chromosome[i+int(context/2)]][0]
							nuc = bias + ":" + nuc

						# Update the dictionary for the current nucleotide
						if nuc not in probs:
							probs[nuc] = {chrom:1}
						else:
							if chrom not in probs[nuc]:
								probs[nuc][chrom] = 1
							else:
								probs[nuc][chrom] += 1

					else:
						if context_input == 'DBS186':
							bias = tsb_ref[chromosome[i+int(context/2)]][0]
							if nuc not in dinuc_types:
								nuc = revcompl(nuc)
								bias = revbias(bias)
							if nuc not in dinuc_tsb:
								bias = 'Q'
							nuc = bias + ":" + nuc

						else:
							if nuc not in dinuc_types:
								nuc = revcompl(nuc)
						# Update the dictionary for the current nucleotide
						if chrom not in probs[nuc]:
							probs[nuc][chrom] = 1
						else:
							probs[nuc][chrom] += 1


						# if context_input == 'DBS186':
						# 	bias = tsb_ref[chromosome[i+int(context/2)]][0]
						# 	nuc = bias + ":" + nuc
						# # Populate the dictionary if desired context is DINUC
						# for dinuc in dinuc_types:
						# 	if dinuc not in probs:
						# 		probs[dinuc] = {}

						# # Update the dictionary for the current nucleotide
						# if nuc not in probs:
						# 	nuc = revcompl(nuc)
							
						# if chrom not in probs[nuc]:
						# 	probs[nuc][chrom] = 1
						# else:
						# 	probs[nuc][chrom] += 1


		print("chrom ", chrom, "done")
	print(probs, chromosome_lengths)
	# Write the resulting dictionary to the csv file
	with open (output_file, 'w') as out:
		print(' ,', end='', file=out)
		for chrom in chromosomes[:-1]:
			print(chrom + ',', end='',file=out)
		print(chromosomes[-1],file=out)
		for nuc in probs.keys():
			nuc_sum = sum(probs[nuc].values())
			print (nuc + ',', end='', file=out)
			for i in range (0, len(chromosomes[:-1]), 1):
				try:
					print(str(probs[nuc][chromosomes[i]]/nuc_sum) + ',', end='', file=out)
					out.flush()
				except:
					print(str(0) + ',', end='', file=out)
			try:
				print(probs[nuc][chromosomes[-1]]/nuc_sum, file=out)
			except:
				print(str(0), file=out)

	counts_file = os.path.dirname(output_file)	
	with open (counts_file + "/context_counts_" + genome + "_" + context_input + ".csv", 'w') as out:
		print(' ,', end='', file=out)
		for chrom in chromosomes[:-1]:
			print(chrom + ',', end='',file=out)
		print(chromosomes[-1],file=out)
		for nuc in probs.keys():
			nuc_sum = sum(probs[nuc].values())
			print (nuc + ',', end='', file=out)
			for chroms in chromosomes[:-1]:
				try:
					print(str(probs[nuc][chroms]) + ',', end='', file=out)
					out.flush()
				except:
					print(str(0) + ',', end='', file=out)
			try:
				print(str(probs[nuc][chroms]), file=out)
			except:
				print(str(0), file=out)
			# try:
			# 	print(probs[nuc][chromosomes[-1]]/nuc_sum, file=out)
			# except:
			# 	print(str(0),file=out)
	# chromosomes.remove("Y")
	# chromosomes.remove("MT")
	# chromosomes.remove("M")

	
	# Sort the file so that the nucleotides are in alphabetical order
	sort_command_1 = "sort -t ',' -k 1,1 "
	sort_command_2 = " -o "
	os.system (sort_command_1 + output_file + sort_command_2 + output_file)


def context_distribution_BED (context_input, output_file, chromosome_path, chromosomes, bed, bed_file, exome, exome_file, genome, ref_dir, tsb_ref, gender):
	'''
	Creates a csv file for the distribution of nucleotides given a specific context and BED file. 
	This csv file needs to be created before simulating mutationalsigantures for the given 
	context.

	Requires: 
		Chromosomes saved in individual text files ('X.txt','Y.txt', '1.txt', etc)
		Transcriptional data saved in binary files for each chromosome. These files
			can be created using the script: "save_tsb_192.py"

	Parameters:
			  context_input  -> simulation context of interest (ex: 96, 192, 1536, 3072, DINUC, INDEL)
				output_file  -> file where the distribution for the given nucleotide context is saved (csv file)
			chromosome_path  -> path to the reference chromosomes
				chromosomes  -> list of chromosomes for the species of interest
						bed  -> flag that determines if the user has provided a BED file with specific ranges to simulate
				   bed_file  -> BED file that contains the ranges of interest

	Returns:
		None

	Outputs:
		CSV file with each chromosome represented as a column and reach row 
		represented as a nucleotide. Under each chromosome for a given nucleotide
		is the proportion of the total length of that chromosome associated with that 
		nucleotide. 
	'''


	dinuc_types = ['AA', 'AC', 'AG', 'AT', 'CA', 'CC', 'CG', 'GA', 'GC', 'TA']
	dinuc_tsb = ['AA', 'AG', 'CC', 'GA',]
	dinuc_non_tsb = ['Q:AC', 'Q:AT', 'Q:CA', 'Q:CG', 'Q:GC', 'Q:TA']
	tsb_bias = ['T','U','B','N']

	# Set the context parameter based upon the user input
	if context_input == '96' or context_input == '192' or context_input == '384':
		context = 3
	elif context_input == '1536' or context_input == '3072' or context_input == '6144':
		context = 5
	elif context_input == 'DINUC' or context_input == 'DBS' or context_input == 'DBS186':
		context = 2
	elif context_input == '6' or context_input == '24':
		context = 1
	else:
		print('Not a valid context')
		sys.exit()
	count = 0 
	probs = {}
	chromosome_lengths = {}
	first_line = True
	chrom_length = 0
	if exome:
		# file_to_open = ref_dir + "/references/chromosomes/exome/" + genome + "/" + exome_file
		file_to_open = exome_file
	else:
		# file_to_open = ref_dir + "/references/vcf_files/BED/" + bed_file
		file_to_open = bed_file

	chromosomes_sort = chromosomes
	if "Y" not in chromosomes:
		chromosomes_sort.append("Y")
	if "MT" not in chromosomes:
		chromosomes_sort.append("MT")
	if "M" not in chromosomes:
		chromosomes_sort.append("M")

	# Populate the dictionary if desired context is DINUC
	if context_input == 'DBS' or context_input == 'DINUC': 
		for dinuc in dinuc_types:
			probs[dinuc] = {}

	if context_input == 'DBS186':
		for dinuc in dinuc_tsb:
			for bias in tsb_bias:
				dinuc_t = bias + ":" + dinuc
				probs[dinuc_t] = {}
		for dinuc in dinuc_non_tsb:
			probs[dinuc] = {}


	with open(file_to_open) as f:
		lines = [line.strip().split() for line in f]
	output = open(file_to_open, 'w')
	print('\t'.join(lines[0]), file=output)
	if len(lines) > 1:
		if len(lines[1][0]) > 2:
			for line in sorted(lines[1:], key = lambda x: (chromosomes_sort.index(x[0][3:]), int(x[1]), int(x[2]))):
				print('\t'.join(line), file=output)
		else:
			for line in sorted(lines[1:], key = lambda x: (chromosomes_sort.index(x[0]), int(x[1]), int(x[2]))):
				print('\t'.join(line), file=output)
	output.close()

	with open(file_to_open) as b_file:
		next(b_file)
		for lines in b_file:
			line = lines.strip().split()
			chrom = line[0]
			if len(chrom) > 1 and chrom[0:3].upper() == 'CHR':
				chrom = chrom[3:]
			if gender == 'female' and chrom == 'Y':
				continue
			start = int(line[1])
			end = int(line[2])

			if first_line:
				chrom_initial = chrom
				first_line = False 
				f = open (chromosome_path + chrom + ".txt", "rb")
				chromosome = f.read()

			if chrom == chrom_initial:
				chrom_length += end-start
				for i in range(start, end+1-context, 1):
					nuc = ''
					for l in range(i, i+context,1):
						nuc += tsb_ref[chromosome[l]][1]
					base = nuc[int(context/2)]

					# Skip the base if unknown 
					if "N" in nuc:
						pass

					else:
						if context_input != "DINUC" and context_input != 'DBS186':
							# Only save the pyrimidine context (canonical)
							if base == 'A' or base == 'G':
								nuc = revcompl(nuc)

							# Adjust the nucleotide representaiton if TSB is desired
							if context_input == '192' or context_input == '3072' or context_input == '384' or context_input == '6144' or context_input == '24':
								bias = tsb_ref[chromosome[i+int(context/2)]][0]
								nuc = bias + ":" + nuc

							# Update the dictionary for the current nucleotide
							if nuc not in probs:
								probs[nuc] = {chrom:1}
							else:
								if chrom not in probs[nuc]:
									probs[nuc][chrom] = 1
								else:
									probs[nuc][chrom] += 1

						else:
							if context_input == 'DBS186':
								bias = tsb_ref[chromosome[i+int(context/2)]][0]
								if nuc not in dinuc_types:
									nuc = revcompl(nuc)
									bias = revbias(bias)
								if nuc not in dinuc_tsb:
									bias = 'Q'
								nuc = bias + ":" + nuc

							else:
								if nuc not in dinuc_types:
									nuc = revcompl(nuc)
							# Update the dictionary for the current nucleotide
							if chrom not in probs[nuc]:
								probs[nuc][chrom] = 1
							else:
								probs[nuc][chrom] += 1
			else:
				f.close()
				print("          Chromosome ", chrom_initial, "done")
				chromosome_lengths[chrom_initial] = chrom_length
				chrom_length = end-start
				chrom_initial = chrom
				try:
					f = open (chromosome_path + chrom + ".txt", "rb")
					chromosome = f.read()
				except:
					continue

				for i in range(start, end+1-context, 1):
					nuc = ''
					for l in range(i,i+context,1):
						nuc += tsb_ref[chromosome[l]][1]
					base = nuc[int(context/2)]

					# Skip the base if unknown 
					if "N" in nuc:
						pass

					else:
						if context_input != "DINUC" and context_input != 'DBS186':
							# Only save the pyrimidine context (canonical)
							if base == 'A' or base == 'G':
								nuc = revcompl(nuc)

							# Adjust the nucleotide representaiton if TSB is desired
							if context_input == '192' or context_input == '3072' or context_input == '384' or context_input == '6144' or context_input == '24':
								bias = tsb_ref[chromosome[i+int(context/2)]][0]
								nuc = bias + ":" + nuc

							# Update the dictionary for the current nucleotide
							if nuc not in probs:
								probs[nuc] = {chrom:1}
							else:
								if chrom not in probs[nuc]:
									probs[nuc][chrom] = 1
								else:
									probs[nuc][chrom] += 1
						else:
							if context_input == 'DBS186':
								bias = tsb_ref[chromosome[i+int(context/2)]][0]
								if nuc not in dinuc_types:
									nuc = revcompl(nuc)
									bias = revbias(bias)
								if nuc not in dinuc_tsb:
									bias = 'Q'
								nuc = bias + ":" + nuc

							else:
								if nuc not in dinuc_types:
									nuc = revcompl(nuc)
							# Update the dictionary for the current nucleotide
							if chrom not in probs[nuc]:
								probs[nuc][chrom] = 1
							else:
								probs[nuc][chrom] += 1
		chromosome_lengths[chrom_initial] = chrom_length

	# Write the resulting dictionary to the csv file
	with open (output_file, 'w') as out:
		print(' ,', end='', file=out)
		for chrom in chromosomes[:-1]:
			print(chrom + ',', end='',file=out)
		print(chromosomes[-1],file=out)
		for nuc in probs.keys():
			nuc_sum = sum(probs[nuc].values())
			print (nuc + ',', end='', file=out)
			for chroms in chromosomes[:-1]:
				try:
					print(str(probs[nuc][chroms]/nuc_sum) + ',', end='', file=out)
					out.flush()
				except:
					print(str(0) + ',', end='', file=out)
			try:
				print(probs[nuc][chromosomes[-1]]/nuc_sum, file=out)
			except:
				print(str(0),file=out)

	counts_file = os.path.dirname(output_file)	
	with open (counts_file + "/context_counts_" + genome + "_" + context_input + "_exome.csv", 'w') as out:
		print(' ,', end='', file=out)
		for chrom in chromosomes[:-1]:
			print(chrom + ',', end='',file=out)
		print(chromosomes[-1],file=out)
		for nuc in probs.keys():
			nuc_sum = sum(probs[nuc].values())
			print (nuc + ',', end='', file=out)
			for chroms in chromosomes[:-1]:
				try:
					print(str(probs[nuc][chroms]) + ',', end='', file=out)
					out.flush()
				except:
					print(str(0) + ',', end='', file=out)
			# try:
			# 	print(probs[nuc][chromosomes[-1]]/nuc_sum, file=out)
			# except:
			# 	print(str(0),file=out)
			try:
				print(str(probs[nuc][chroms]), file=out)
			except:
				print(str(0), file=out)
	# chromosomes.remove("Y")
	# chromosomes.remove("MT")
	# chromosomes.remove("M")

	# Sort the file so that the nucleotides are in alphabetical order
	sort_command_1 = "sort -t ',' -k 1,1 "
	sort_command_2 = " -o "
	os.system (sort_command_1 + output_file + sort_command_2 + output_file)




def main():
	bed = False
	bed_file = None
	exome = False
	exome_file = None
	gender = 'male'

	parser = argparse.ArgumentParser(description="Provide the necessary arguments to save the nucleotide distributions for each chromosome.")
	parser.add_argument("--genome", "-g",help="Provide a reference genome. (ex: GRCh37, GRCh38, mm10)")
	parser.add_argument("--context", "-c",help="Whole genome context by default")
	parser.add_argument("-b", "--bed", nargs='?', help="Optional parameter instructs script to simulate on a given set of ranges (ex: exome). Whole genome context by default")
	parser.add_argument("-e", "--exome", nargs='?', help="Optional parameter instructs script to simulate only on exome). Whole genome context by default")
	parser.add_argument("-gD", "--gender", help="Optional parameter instructs script to create the context files based on female (two x chromosomes.", action='store_true')

	tsb_ref = {0:['N','A'], 1:['N','C'], 2:['N','G'], 3:['N','T'],
			   4:['T','A'], 5:['T','C'], 6:['T','G'], 7:['T','T'],
			   8:['U','A'], 9:['U','C'], 10:['U','G'], 11:['U','T'],
			   12:['B','A'], 13:['B','C'], 14:['B','G'], 15:['B','T'],
			   16:['N','N'], 17:['T','N'], 18:['U','N'], 19:['B','N']}

	args=parser.parse_args()
	genome = args.genome
	context = args.context
	if args.bed:
		bed = True
		bed_file = args.bed

	if args.exome:
		exome = True
		exome_file = args.exome

	chromosomes = ['X', 'Y', '1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', 
				   '13', '14', '15', '16', '17', '18', '19', '20', '21', '22']

	if args.gender:
		gender = 'female'
		chromosomes.remove('Y')


	if genome.upper() == 'MM10' or genome.upper() == 'MM9' or genome.upper() == 'MM39':
		chromosomes = chromosomes[:21]
	if genome.upper() == 'RN6':
		chromosomes = chromosomes[:22]

	script_dir = os.getcwd()
	ref_dir = re.sub('\/scripts$', '', script_dir)
	#chromosome_path = ref_dir + "/references/chromosomes/chrom_string/" + genome + "/"
	#chromosome_path = "/Users/ebergstr/Desktop/test_bi/references/chromosomes/tsb/GRCh37/"
	chromosome_path = ref_dir + "/references/chromosomes/tsb/" + genome + "/"

	output_path = ref_dir + '/references/chromosomes/context_distributions/'
	if not os.path.exists(output_path):
		os.makedirs(output_path)

	if bed:
		output_file = ref_dir + '/references/chromosomes/context_distributions/context_distribution_' + genome + "_" + context + "_" + gender + '_BED.csv'
	else:
		if exome:
			output_file = ref_dir + '/references/chromosomes/context_distributions/context_distribution_' + genome + "_" + context + "_" + gender +'_exome.csv'
		else:
			output_file = ref_dir + '/references/chromosomes/context_distributions/context_distribution_' + genome + "_" + context + "_" + gender +'.csv'

	if bed or exome:
		context_distribution_BED(context, output_file, chromosome_path, chromosomes, bed, bed_file, exome, exome_file, genome, ref_dir, tsb_ref)
	else:
		context_distribution(context, output_file, chromosome_path, chromosomes, tsb_ref, genome)
	
if __name__ == '__main__':
	main()




import os
import shutil
import subprocess
import tempfile
import re

import pysam

import handygenome
import handygenome.workflow as workflow
import handygenome.variant.variantplus as variantplus
import handygenome.bameditor as bameditor


GATK = os.path.join(handygenome.DIRS['externals'], 'gatk_wrapper.sh')


def run_haplotypecaller(fasta_path, infile_path, outfile_path, tmpdir=None, 
                        incl_region_path=None, excl_region_path=None,
                        mem_gb=6, threads=4, mbq=10, rm_tmpdir=True):
    """Args:
        threads: value for --native-pair-hmm-threads option. 
            gatk default is 4.
    """
    if tmpdir is None:
        tmpdir = tempfile.mkdtemp(prefix='HaplotypeCaller_tmpdir_', dir=os.getcwd())

    args = [
        GATK, 
        '--java-options', f'-Xmx{mem_gb}G',
        'HaplotypeCaller',
        '--reference', fasta_path,
        '--input', infile_path,
        '--output', outfile_path,
        '--native-pair-hmm-threads', str(threads),
        '--min-base-quality-score', str(mbq),
        '--sample-ploidy', '2', # gatk default
        '--active-probability-threshold', '0.002', # gatk default
        '--standard-min-confidence-threshold-for-calling', '30', # gatk default
        '--dont-use-soft-clipped-bases', 'false',
        '--create-output-variant-index', 'false',
        '--tmp-dir', tmpdir,
        ]

    if incl_region_path is not None:
        args.extend(['--intervals', incl_region_path])
    if excl_region_path is not None:
        args.extend(['--exclude-intervals', excl_region_path])

    p = subprocess.run(args, capture_output=True, text=True, check=False, shell=False) 
    if rm_tmpdir:
        shutil.rmtree(tmpdir)

    return p


def run_Mutect2(infile_path_list, outfile_path, fasta_path, tmpdir=None, realigned_bam_path=None, incl_region_path=None, excl_region_path=None, mem_gb=12, threads=4, mbq=10, rm_tmpdir=True):
    """For 4.3.0.0"""

    if tmpdir is None:
        tmpdir = tempfile.mkdtemp(prefix='Mutect2_tmpdir_', dir=os.getcwd())

    args = [
        GATK, 
        '--java-options', f'-Xmx{mem_gb}G',
        'Mutect2',
        '--reference', fasta_path,
        #'--input', infile_path,
        '--output', outfile_path,

        '--force-active', 'true',

        '--min-base-quality-score', str(mbq),
        '--native-pair-hmm-threads', str(threads),
        '--active-probability-threshold', '0.002', # gatk default
        '--dont-use-soft-clipped-bases', 'false',
        '--create-output-variant-index', 'false',
        '--tmp-dir', tmpdir,

        '--min-assembly-region-size', f'{50}',  # gatk default
        '--assembly-region-padding', f'{100}',
        #'--dont-trim-active-regions', 'true',
    ]

    for infile_path in infile_path_list:
        args.extend(['--input', infile_path])

    if realigned_bam_path is not None:
        args.extend(['--bam-output', realigned_bam_path])

    if incl_region_path is not None:
        args.extend(['--intervals', incl_region_path])
    if excl_region_path is not None:
        args.extend(['--exclude-intervals', excl_region_path])

    p = subprocess.run(args, capture_output=True, text=True, check=False, shell=False) 
    if rm_tmpdir:
        shutil.rmtree(tmpdir)

    return p


def realign_with_Mutect2(bam_path_list, chrom, start0, end0, fasta_path, realigned_bam_path):
    tmpdir = tempfile.mkdtemp(prefix='tmpdir_Mutect2_realignment_', dir=os.getcwd())
    out_vcf_path = os.path.join(tmpdir, 'output.vcf.gz')

    region_path = os.path.join(tmpdir, 'region.bed')
    with open(region_path, 'wt') as infile:
        infile.write(f'{chrom}\t{start0}\t{end0}\n')

    p = run_Mutect2(bam_path_list, out_vcf_path, fasta_path, tmpdir=tmpdir, realigned_bam_path=realigned_bam_path, incl_region_path=region_path, rm_tmpdir=False)
    if p.returncode != 0:
        raise Exception(p.stderr)

    result_vplist = variantplus.load_vcf(out_vcf_path)

    shutil.rmtree(tmpdir)

    return result_vplist


def split_realigned_bam(realigned_bam_path):
    in_bam = pysam.AlignmentFile(realigned_bam_path, mode='rb')
    readgroups = [x['ID'] for x in in_bam.header['RG'] if x['ID'] != 'ArtificialHaplotypeRG']
    out_bam_paths = {
        rg: re.sub('.bam$', f'.{rg}.bam', realigned_bam_path)
        for rg in readgroups
    }
    out_bams = {
        rg: pysam.AlignmentFile(out_bam_paths[rg], mode='wb', header=in_bam.header)
        for rg in readgroups
    }

    for read in in_bam.fetch():
        rg_tag = read.get_tag('RG')
        if rg_tag != 'ArtificialHaplotypeRG':
            out_bams[read.get_tag('RG')].write(read)

    for bam in out_bams.values():
        bam.close()
    for bam_path in out_bam_paths.values():
        bameditor.sort_and_index(bam_path)

    split_bams = {
        rg: pysam.AlignmentFile(out_bam_paths[rg], mode='rb')
        for rg in readgroups
    }
    return split_bams



import os
import itertools

import pysam

import handygenome.interval as libinterval
import handygenome.workflow as workflow
import handygenome.logutils as logutils
import handygenome.vcfeditor.misc as vcfmisc
import handygenome.deco as deco


LOGGER = logutils.get_logger()


def sanity_check(vcf_path, outdir):
    vcf_path = workflow.arghandler_infile(vcf_path)
    outdir = workflow.arghandler_outdir(outdir)

    return vcf_path, outdir


@deco.get_deco_num_set_differently(('n_file', 'n_line'), 1)
def get_output_lineno_list(total_lineno, n_file=None, n_line=None):
    """
    Args:
        total_lineno: Total number of variant records of input vcf.
        n_file: Number of output files. If greater than the total line number, 
            reduced to the total line number.
        n_line: Number of variant records per one output file. If greater 
            than the total line number, reduced to the total line number.
        * One and only one of n_file or n_line argument must be set.

    Returns:
        A list composed of the number of variant records per output file.
    """

    def warn():
        LOGGER.warning(
            f'"n_file" is greater than "total_lineno". It will be '
            f'changed to be the same with "total_lineno".')
    
    assert total_lineno > 0, f'"total_lineno" must be a positive integer.'

    if n_file is not None:
        if n_file > total_lineno:
            warn()
        result = libinterval.get_interval_lengths_num(total_lineno, n_file)
    elif n_line is not None:
        if n_line > total_lineno:
            warn()
        result = libinterval.get_interval_lengths_width(total_lineno, n_line)

    return result


def write_split_vcfs(vcf_path, output_lineno_list, split_filenames, 
                     mode_pysam):
    with pysam.VariantFile(vcf_path, 'r') as in_vcf:
        fetch = in_vcf.fetch()
        for lineno, outfile_path in zip(output_lineno_list, split_filenames):
            with pysam.VariantFile(outfile_path, mode_pysam, 
                                   header=in_vcf.header) as out_vcf:
                for _ in range(lineno):
                    out_vcf.write(next(fetch))


@deco.get_deco_num_set_differently(('n_file', 'n_line'), 1)
def main(vcf_path, outdir, n_file=None, n_line=None, mode_bcftools='z', 
         mode_pysam=None, prefix='', suffix='.vcf.gz'):
    """
    Splits input vcf by either 'n_file' (number of output files) or 
        'n_line' (number of lines each split file contains).
    Split file names looks like: <prefix>000<suffix>

    Args:
        vcf_path: Input vcf file path
        outdir: Directory where split vcf files will be saved. May not be 
            existing. Must not be an non-empty directory.
        One of 'n_file' and 'n_line' must be set.
    """

    vcf_path, outdir = sanity_check(vcf_path, outdir)

    total_lineno = vcfmisc.get_vcf_lineno(vcf_path)
    output_lineno_list = get_output_lineno_list(
            total_lineno, n_file = n_file, n_line = n_line,
            )
    mode_pysam = vcfmisc.write_mode_arghandler(mode_bcftools, mode_pysam)
    split_filenames = workflow.get_split_filenames( 
        len(output_lineno_list), outdir, prefix, suffix)
        # "split_filenames" begins with 0

    write_split_vcfs(vcf_path, output_lineno_list, split_filenames, 
                     mode_pysam)

    return split_filenames



import pyranges as pr

import handygenome.annotation.data as annotdata
import handygenome.annotation.annotitem as libannotitem
import handygenome.analysis.feature as anal_feature


def gene_cosmic_occurrences(gene_names, refver):
    result = dict()
    cosmic_vcf = annotdata.VCFS_COSMIC[refver]
    # print(f'getting coordinates for each gene')
    gene_coords = anal_feature.get_gene_coords(gene_names, refver)
    for gene_name, interval in gene_coords.items():
        # print(f'counting occurrence for gene {gene_name}')
        occurrence = 0
        for vr in cosmic_vcf.fetch(interval.chrom, interval.start0, interval.end0):
            cosmic_info = libannotitem._decode(vr.info['cosmic_info'][0])
            occurrence += sum(cosmic_info['occurrence'].values())
        result[gene_name] = occurrence
    
    return result
